from pathlib import Path
from tifffile import TiffFile
from wsireg.reg_images import (
    CziRegImage,
    SitkRegImage,
    NumpyRegImage,
    # AICSRegImage,
)
from wsireg.utils.im_utils import (
    TIFFFILE_EXTS,
    ARRAYLIKE_CLASSES,
    tifffile_to_arraylike,
    ome_tifffile_to_arraylike,
)


def reg_image_loader(
    image,
    image_res,
    mask=None,
    pre_reg_transforms=None,
    preprocessing=None,
    channel_names=None,
    channel_colors=None,
):
    if isinstance(image, ARRAYLIKE_CLASSES):
        return NumpyRegImage(
            image,
            image_res,
            mask=mask,
            pre_reg_transforms=pre_reg_transforms,
            preprocessing=preprocessing,
            channel_names=channel_names,
            channel_colors=channel_colors,
            image_filepath=None,
        )

    image_ext = Path(image).suffix
    if image_ext in TIFFFILE_EXTS:
        if TiffFile(image).is_ome:
            image, image_filepath = ome_tifffile_to_arraylike(image)
            reg_image = NumpyRegImage(
                image,
                image_res,
                mask=mask,
                pre_reg_transforms=pre_reg_transforms,
                preprocessing=preprocessing,
                channel_names=channel_names,
                channel_colors=channel_colors,
                image_filepath=image_filepath,
            )
        else:
            image, image_filepath = tifffile_to_arraylike(image)
            reg_image = NumpyRegImage(
                image,
                image_res,
                mask=mask,
                pre_reg_transforms=pre_reg_transforms,
                preprocessing=preprocessing,
                channel_names=channel_names,
                channel_colors=channel_colors,
                image_filepath=image_filepath,
            )
    elif image_ext == ".czi":
        reg_image = CziRegImage(
            image,
            image_res,
            mask=mask,
            pre_reg_transforms=pre_reg_transforms,
            preprocessing=preprocessing,
            channel_names=channel_names,
            channel_colors=channel_colors,
        )
    else:
        reg_image = SitkRegImage(
            image,
            image_res,
            mask=mask,
            pre_reg_transforms=pre_reg_transforms,
            preprocessing=preprocessing,
            channel_names=channel_names,
            channel_colors=channel_colors,
        )

    return reg_image



BASE_RIG_TFORM = dict(
    {
        "Transform": ["EulerTransform"],
        "NumberOfParameters": ["3"],
        "TransformParameters": ["0", "0", "0"],
        "InitialTransformParametersFileName": ["NoInitialTransform"],
        "HowToCombineTransforms": ["Compose"],
        "FixedImageDimension": ["2"],
        "MovingImageDimension": ["2"],
        "FixedInternalImagePixelType": ["float"],
        "MovingInternalImagePixelType": ["float"],
        "Size": ["0", "0"],
        "Index": ["0", "0"],
        "Spacing": ["", ""],
        "Origin": ["0.0000", "0.0000"],
        "Direction": [
            "1.0000000000",
            "0.0000000000",
            "0.0000000000",
            "1.0000000000",
        ],
        "UseDirectionCosines": ["true"],
        "CenterOfRotationPoint": ["0", "0"],
        "ResampleInterpolator": ["FinalNearestNeighborInterpolator"],
        "Resampler": ["DefaultResampler"],
        "DefaultPixelValue": ["0.000000"],
        "ResultImageFormat": ["mha"],
        "ResultImagePixelType": ["float"],
        "CompressResultImage": ["true"],
    }
)

BASE_TRANSLATION_TFORM = dict(
    {
        "Transform": ["TranslationTransform"],
        "NumberOfParameters": ["2"],
        "TransformParameters": ["0", "0"],
        "InitialTransformParametersFileName": ["NoInitialTransform"],
        "HowToCombineTransforms": ["Compose"],
        "FixedImageDimension": ["2"],
        "MovingImageDimension": ["2"],
        "FixedInternalImagePixelType": ["float"],
        "MovingInternalImagePixelType": ["float"],
        "Size": ["0", "0"],
        "Index": ["0", "0"],
        "Spacing": ["", ""],
        "Origin": ["0.0000", "0.0000"],
        "Direction": [
            "1.0000000000",
            "0.0000000000",
            "0.0000000000",
            "1.0000000000",
        ],
        "UseDirectionCosines": ["true"],
        "CenterOfRotationPoint": ["0", "0"],
        "ResampleInterpolator": ["FinalNearestNeighborInterpolator"],
        "Resampler": ["DefaultResampler"],
        "DefaultPixelValue": ["0.000000"],
        "ResultImageFormat": ["mha"],
        "ResultImagePixelType": ["float"],
        "CompressResultImage": ["true"],
    }
)


BASE_AFF_TFORM = dict(
    {
        "Transform": ["AffineTransform"],
        "NumberOfParameters": ["6"],
        "TransformParameters": ["1", "0", "0", "1", "0", "0"],
        "InitialTransformParametersFileName": ["NoInitialTransform"],
        "HowToCombineTransforms": ["Compose"],
        "FixedImageDimension": ["2"],
        "MovingImageDimension": ["2"],
        "FixedInternalImagePixelType": ["float"],
        "MovingInternalImagePixelType": ["float"],
        "Size": ["0", "0"],
        "Index": ["0", "0"],
        "Spacing": ["0", "0"],
        "Origin": ["0.0000", "0.0000"],
        "Direction": [
            "1.0000000000",
            "0.0000000000",
            "0.0000000000",
            "1.0000000000",
        ],
        "UseDirectionCosines": ["true"],
        "CenterOfRotationPoint": ["0", "0"],
        "ResampleInterpolator": ["FinalNearestNeighborInterpolator"],
        "Resampler": ["DefaultResampler"],
        "DefaultPixelValue": ["0.000000"],
        "ResultImageFormat": ["mha"],
        "ResultImagePixelType": ["float"],
        "CompressResultImage": ["true"],
    }
)



from copy import deepcopy
from pathlib import Path
import json
import numpy as np
import SimpleITK as sitk
import cv2
import geojson
from wsireg.reg_transform import RegTransform
from wsireg.utils.tform_utils import (
    wsireg_transforms_to_itk_composite,
)

GJ_SHAPE_TYPE = {
    "polygon": geojson.Polygon,
    "multipolygon": geojson.MultiPolygon,
    "point": geojson.Point,
    "multipoint": geojson.MultiPoint,
    "multilinestring": geojson.MultiLineString,
    "linestring": geojson.LineString,
}


def gj_to_np(gj: dict):
    """
    Convert geojson representation to np.ndarray representation of shape

    Parameters
    ----------
    gj : dict
        GeoJSON data stored as python dict

    Returns
    -------
    dict
        containing keys
            "array": np.ndarray - x,y point data in array
            "shape_type": str - indicates GeoJSON shape_type (Polygon, MultiPolygon, etc.)
            "shape_name": str - name inherited from QuPath GeoJSON
    """
    if gj.get("geometry").get("type") == "MultiPolygon":
        pts = []
        for geo in gj.get("geometry").get("coordinates"):
            pts.append(np.squeeze(np.array(geo)))
        pts = np.vstack(pts)

    elif gj.get("geometry").get("type") == "Polygon":
        pts = np.squeeze(np.asarray(gj.get("geometry").get("coordinates")))
    elif gj.get("geometry").get("type") == "Point":
        pts = np.expand_dims(
            np.asarray(gj.get("geometry").get("coordinates")), 0
        )
    elif gj.get("geometry").get("type") == "MultiPoint":
        pts = np.asarray(gj.get("geometry").get("coordinates"))
    elif gj.get("geometry").get("type") == "LineString":
        pts = np.asarray(gj.get("geometry").get("coordinates"))

    if gj.get("properties").get("classification") is None:
        shape_name = "unnamed"
    else:
        shape_name = gj.get("properties").get("classification").get("name")

    return {
        "array": pts.astype(np.double),
        "shape_type": gj.get("geometry").get("type"),
        "shape_name": shape_name,
    }


def add_unamed(gj):
    if gj.get("properties").get("classification") is None:
        gj.get("properties").update({"classification": {"name": "unnamed"}})
    return gj


def read_geojson(json_file: str):
    """Read GeoJSON files (and some QuPath metadata).

    Parameters
    ----------
    json_file : str
        file path of QuPath exported GeoJSON
    Returns
    -------
    gj_data : dict
        dict of GeoJSON information
    shapes_np : dict
        dict of GeoJSON information stored in np.ndarray
            "array": np.ndarray - x,y point data in array
            "shape_type": str - indicates GeoJSON shape_type (Polygon, MultiPolygon, etc.)
            "shape_name": str - name inherited from QuPath GeoJSON
    """
    gj_data = json.load(open(json_file, "r"))

    shapes_np = [gj_to_np(s) for s in gj_data]
    gj_data = [add_unamed(gj) for gj in gj_data]
    return gj_data, shapes_np


def np_to_geojson(
    np_array: np.ndarray, shape_type="polygon", shape_name="unnamed"
):
    """convert np.ndarray to GeoJSON dict

    Parameters
    ----------
    np_array: np.ndarray
        coordinates of data
    shape_type:str
        GeoJSON shape type
    shape_name:str
        Name of the shape
    Returns
    -------
    shape_gj : dict
        dict of GeoJSON information
    shape_np : dict
        dict of GeoJSON information stored in np.ndarray
            "array": np.ndarray - x,y point data in array
            "shape_type": str - indicates GeoJSON shape_type (Polygon, MultiPolygon, etc.)
            "shape_name": str - name inherited from QuPath GeoJSON
    """
    sh_type = shape_type.lower()

    gj_func = GJ_SHAPE_TYPE[sh_type]

    shape_gj = {
        "geometry": gj_func(np_array.tolist()),
        "properties": {"classification": {"name": shape_name}},
    }
    shape_np = {
        "array": np_array,
        "shape_type": shape_type.lower(),
        "shape_name": shape_name,
    }
    return shape_gj, shape_np


def shape_reader(shape_data, **kwargs):
    """
    Read shape data for transformation
    Shape data is stored as numpy arrays for operations but also as GeoJSON
    to contain metadata and interface with QuPath

    Parameters
    ----------
    shape_data: list of np.ndarray or str
        if str, will read data as GeoJSON file, if np.ndarray with assume
        it is coordinates
    kwargs
        keyword args passed to np_to_geojson convert

    Returns
    -------
    shapes_gj: list of dicts
        list of dicts of GeoJSON information
    shapes_np: list of dicts
        list of dicts of GeoJSON information stored in np.ndarray
        "array": np.ndarray - x,y point data in array
        "shape_type": str - indicates GeoJSON shape_type (Polygon, MultiPolygon, etc.)
        "shape_name": str - name inherited from QuPath GeoJSON
    """
    if isinstance(shape_data, list) is False:
        shape_data = [shape_data]

    shapes_gj = []
    shapes_np = []
    for sh in shape_data:
        if isinstance(sh, dict):
            out_shape_gj, out_shape_np = np_to_geojson(
                sh["array"], sh["shape_type"], sh["shape_name"]
            )

        elif isinstance(sh, np.ndarray):
            out_shape_gj, out_shape_np = np_to_geojson(sh, **kwargs)

        else:
            if Path(sh).is_file():
                sh_fp = Path(sh)

                if sh_fp.suffix in [".json", ".geojson"]:
                    out_shape_gj, out_shape_np = read_geojson(str(sh_fp))
                # elif sh_fp.suffix == ".cz":
                #     out_shape_gj = read_zen_shapes(str(sh_fp))
                #     out_shape_np = [gj_to_np(s) for s in out_shape_gj]
                else:
                    raise ValueError(
                        "{} is not a geojson or numpy array".format(str(sh_fp))
                    )
        if isinstance(out_shape_gj, list):
            shapes_gj.extend(out_shape_gj)
        else:
            shapes_gj.append(out_shape_gj)

        if isinstance(out_shape_np, list):
            shapes_np.extend(out_shape_np)
        else:
            shapes_np.append(out_shape_np)

    return shapes_gj, shapes_np


def scale_shape_coordinates(poly: dict, scale_factor: float):
    """
    Scale coordinates by a factor

    Parameters
    ----------
    poly: dict
        dict of coordinate data contain np.ndarray in "array" key
    scale_factor: float
        isotropic scaling factor for the coordinates

    Returns
    -------
    poly: dict
        dict containing coordinates scaled by scale_factor
    """
    poly_coords = poly["array"]
    poly_coords = poly_coords * scale_factor
    poly["array"] = poly_coords
    return poly


def invert_nonlinear_transforms(itk_transforms: list):
    """
    Check list of sequential ITK transforms for non-linear (i.e., bspline) transforms
    Transformations need to be inverted to transform from moving to fixed space as transformations
    are mapped from fixed space to moving.
    This will first convert any non-linear transforms to a displacement field then invert the displacement field
    using ITK methods. It usually works quite well but is not an exact solution.
    Linear transforms can be inverted on  the fly when transforming points

    Parameters
    ----------
    itk_transforms:list
        list of itk.Transform

    Returns
    -------
    itk_transforms:list
        list of itk.Transform where any non-linear transforms are replaced with an inverted displacement field
    """
    tform_linear = [t.is_linear for t in itk_transforms]

    if all(tform_linear):
        return itk_transforms
    else:
        nl_idxs = np.where(np.array(tform_linear) == 0)[0]
        for nl_idx in nl_idxs:
            if itk_transforms[nl_idx].inverse_transform is None:
                print(
                    f"transform at index {nl_idx} is non-linear and the inverse has not been computed\n"
                    "inverting displacement field(s)...\n"
                    "this can take some time"
                )
                itk_transforms[nl_idx].compute_inverse_nonlinear()

    return itk_transforms


def prepare_pt_transformation_data(transformations, compute_inverse=True):
    """
    Read and prepare wsireg transformation data for point set transformation

    Parameters
    ----------
    transformations
        list of dict containing elastix transformation data or str to wsireg .json file containing
        elastix transformation data
    compute_inverse : bool
        whether or not to compute the inverse transformation for moving to fixed point transformations
    Returns
    -------
    itk_pt_transforms:list
        list of transformation data ready to operate on points
    target_res:
        physical spacing of the final transformation in the transform sequence
        This is needed to map coordinates defined as pixel indices to physical coordinates and then back
    """
    if all([isinstance(t, RegTransform) for t in transformations]) is False:
        composite, transformations = wsireg_transforms_to_itk_composite(
            transformations
        )
    if compute_inverse:
        transformations = invert_nonlinear_transforms(transformations)
    target_res = float(transformations[-1].output_spacing[0])
    return transformations, target_res


def itk_transform_pts(
    pt_data: np.ndarray,
    itk_transforms: list,
    px_idx=True,
    source_res=1,
    output_idx=True,
    target_res=2,
):
    """
    Transforms x,y points stored in np.ndarray using list of ITK transforms
    All transforms are in physical coordinates, so all points must be converted to physical coordinates
    before transformation, but this function allows converting back to pixel indices after transformation

    Can intake points in physical coordinates is px_idx == False
    Can output points in physical coordinates if output_idx == False

    Parameters
    ----------
    pt_data : np.ndarray
        array where rows are points and columns are x,y
    itk_transforms: list
        list of ITK transforms, non-linear transforms should be inverted
    px_idx: bool
        whether points are specified in physical coordinates (i.e., microns) or
        in pixel indices
    source_res: float
        resolution of the image on which annotations were made
    output_idx: bool
        whether transformed points should be output in physical coordinates (i.e., microns) or
        in pixel indices
    target_res: float
        resolution of the final target image for conversion back to pixel indices

    Returns
    -------
    tformed_pts:np.ndarray
        transformed points array where rows are points and columns are x,y

    """
    tformed_pts = []
    for pt in pt_data:
        if px_idx is True:
            pt = pt * source_res
        for idx, t in enumerate(itk_transforms):
            if idx == 0:
                t_pt = t.inverse_transform.TransformPoint(pt)
            else:
                t_pt = t.inverse_transform.TransformPoint(t_pt)
        t_pt = np.array(t_pt)
        if output_idx is True:
            t_pt *= 1 / target_res
        tformed_pts.append(t_pt)

    return np.stack(tformed_pts)


def transform_shapes(
    shape_data: list,
    itk_transforms: list,
    px_idx=True,
    source_res=1,
    output_idx=True,
    target_res=2,
):
    """
    Convenience function to apply itk_transform_pts to a list of shape data

    Parameters
    ----------
    shape_data:
        list of arrays where rows are points and columns are x,y
    itk_transforms: list
        list of ITK transforms, non-linear transforms should be inverted
    px_idx: bool
        whether points are specified in physical coordinates (i.e., microns) or
        in pixel indices
    source_res: float
        resolution of the image on which annotations were made
    output_idx: bool
        whether transformed points should be output in physical coordinates (i.e., microns) or
        in pixel indices
    target_res: float
        resolution of the final target image for conversion back to pixel indices

    Returns
    -------
        transformed_shape_data:list
            list of transformed np.ndarray data where rows are points and columns are x,y
    """
    transformed_shape_data = []
    for sh in shape_data:
        t_ptset = deepcopy(sh)
        ptset = sh.get("array")
        t_pts = itk_transform_pts(
            ptset,
            itk_transforms,
            px_idx=px_idx,
            source_res=source_res,
            output_idx=output_idx,
            target_res=target_res,
        )
        t_ptset["array"] = t_pts
        transformed_shape_data.append(t_ptset)

    return transformed_shape_data


def insert_transformed_pts_gj(gj_data: list, np_data: list):
    """
    insert point data into a list of geojson data

    Parameters
    ----------
    shape_gj : dict
        dict of GeoJSON information
    shape_np : dict
        transformed point data in wsireg shape dict

    Returns
    -------
    shape_gj : dict
        dict of GeoJSON information with updated coordinate information
    """

    gj_data_t = deepcopy(gj_data)
    for sh, gj in zip(np_data, gj_data_t):
        shape_type = gj.get("geometry").get("type")

        if shape_type == "Polygon":
            gj.get("geometry").update({"coordinates": [sh["array"].tolist()]})
        elif shape_type == "Point":
            gj.get("geometry").update(
                {"coordinates": np.squeeze(sh["array"]).tolist()}
            )
        elif shape_type == "MultiPoint":
            gj.get("geometry").update({"coordinates": sh["array"].tolist()})
        elif shape_type == "LineString":
            gj.get("geometry").update({"coordinates": sh["array"].tolist()})

    return gj_data_t


def get_int_dtype(value: int):
    """
    Determine appropriate bit precision for indexed image

    Parameters
    ----------
    value:int
        number of shapes

    Returns
    -------
    dtype:np.dtype
        apppropriate data type for index mask
    """
    if value <= np.iinfo(np.uint8).max:
        return np.uint8
    if value <= np.iinfo(np.uint16).max:
        return np.uint16
    if value <= np.iinfo(np.uint32).max:
        return np.int32
    else:
        raise ValueError("Too many shapes")


def get_all_shape_coords(shapes: list):
    return np.vstack(
        [np.squeeze(sh["geometry"]["coordinates"][0]) for sh in shapes]
    )


# code below is for managing transforms as masks rather than point sets
# will probably not reimplement, if segmentation data can is expressed
# as a mask, it can be transformed as an image (using attachment_modality)
def approx_polygon_contour(mask: np.ndarray, percent_arc_length=0.01):
    """
    Approximate binary mask contours to polygon vertices using cv2.

    Parameters
    ----------
    mask : numpy.ndarray
        2-d numpy array of datatype np.uint8.
    percent_arc_length : float
        scaling of epsilon for polygon approximate vertices accuracy.
        maximum distance of new vertices from original.

    Returns
    -------
    numpy.ndarray
        returns an 2d array of vertices, rows: points, columns: y,x

    """
    contours, hierarchy = cv2.findContours(
        mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE
    )

    if len(contours) > 1:
        contours = [contours[np.argmax([cnt.shape[0] for cnt in contours])]]

    epsilon = percent_arc_length * cv2.arcLength(contours[0], True)
    approx = cv2.approxPolyDP(contours[0], epsilon, True)
    return np.squeeze(approx).astype(np.uint32)


def index_mask_to_shapes(index_mask, shape_name, tf_shapes):
    """
    Find the polygons of a transformed shape mask, conveting binary mask
    to list of polygon verteces and sorting by numerical index

    Parameters
    ----------
    index_mask:np.ndarray
        mask where each shape is defined by it's index
    shape_name:str
        name of the shape
    tf_shapes:list
        original list of shape GeoJSON data to be updated

    Returns
    -------
    updated_shapes:list
        dict of GeoJSON information with updated coordinate information
    """
    labstats = sitk.LabelShapeStatisticsImageFilter()
    labstats.SetBackgroundValue(0)
    labstats.Execute(index_mask)

    index_mask = sitk.GetArrayFromImage(index_mask)
    updated_shapes = deepcopy(tf_shapes)

    for idx, shape in enumerate(tf_shapes):
        if shape["properties"]["classification"]["name"] == shape_name:
            label_bb = labstats.GetBoundingBox(idx + 1)
            x_min = label_bb[0]
            x_len = label_bb[2]
            y_min = label_bb[1]
            y_len = label_bb[3]

            sub_mask = index_mask[y_min : y_min + y_len, x_min : x_min + x_len]

            sub_mask[sub_mask == idx + 1] = 255

            yx_coords = approx_polygon_contour(sub_mask, 0.00001)
            xy_coords = yx_coords
            xy_coords = np.append(xy_coords, xy_coords[:1, :], axis=0)
            xy_coords = xy_coords + [x_min, y_min]
            updated_shapes[idx]["geometry"]["coordinates"] = [
                xy_coords.tolist()
            ]

    return updated_shapes


# don't intend to maintain
# def read_zen_shapes(zen_fp):
#     """Read Zeiss Zen Blue .cz ROIs files to wsimap shapely format.
#
#     Parameters
#     ----------
#     zen_fp : str
#         file path of Zen .cz.
#
#     Returns
#     -------
#     list
#         list of wsimap shapely rois
#
#     """
#
#     root = etree.parse(zen_fp)
#
#     rois = root.xpath("//Elements")[0]
#     shapes_out = []
#     for shape in rois:
#         try:
#             ptset_name = shape.find("Attributes/Name").text
#         except AttributeError:
#             ptset_name = "unnamed"
#
#         if shape.tag == "Polygon":
#             ptset_cz = shape.find("Geometry/Points")
#             # ptset_type = "Polygon"
#
#             poly_str = ptset_cz.text
#             poly_str = poly_str.split(" ")
#             poly_str = [poly.split(",") for poly in poly_str]
#             poly_str = [[float(pt[0]), float(pt[1])] for pt in poly_str]
#
#             poly = {
#                 "geometry": geojson.Polygon(poly_str),
#                 "properties": {"classification": {"name": ptset_name}},
#             }
#
#             shapes_out.append(poly)
#
#         if shape.tag == "Rectangle":
#             rect_pts = shape.find("Geometry")
#
#             x = float(rect_pts.findtext("Left"))
#             y = float(rect_pts.findtext("Top"))
#             width = float(rect_pts.findtext("Width"))
#             height = float(rect_pts.findtext("Height"))
#
#             rect = geojson.Polygon(
#                 [
#                     [x, y],
#                     [x + width, y],
#                     [x + width, y + height],
#                     [x, y + height],
#                     [x, y],
#                 ]
#             )
#
#             rect = {
#                 "geometry": rect,
#                 "properties": {"classification": {"name": ptset_name}},
#             }
#
#             shapes_out.append(rect)
#
#     return shapes_out



from pathlib import Path
import numpy as np
from tifffile import TiffWriter
import cv2
import SimpleITK as sitk
from wsireg.writers.ome_tiff_writer import OmeTiffWriter
from wsireg.reg_images.merge_reg_image import MergeRegImage
from wsireg.utils.im_utils import (
    transform_plane,
)
from wsireg.utils.tform_utils import (
    prepare_wsireg_transform_data,
)


class MergeOmeTiffWriter(OmeTiffWriter):
    def __init__(self, reg_image: MergeRegImage):
        super().__init__(reg_image)

    def merge_write_image_by_plane(
        self,
        image_name,
        sub_image_names,
        transformations=None,
        output_dir="",
        write_pyramid=True,
        tile_size=512,
        compression="default",
    ):

        if isinstance(sub_image_names, list) is False:
            if sub_image_names is None:
                sub_image_names = [
                    "" for i in range(len(self.reg_image.images))
                ]
            else:
                raise ValueError(
                    "MergeRegImage requires a list of image names for each image to merge"
                )

        if isinstance(transformations, list) is False:
            if transformations is None:
                transformations = [
                    None for i in range(len(self.reg_image.images))
                ]
            else:
                raise ValueError(
                    "MergeRegImage requires a list of image transforms for each image to merge"
                )

        if len(transformations) != len(self.reg_image.images):
            raise ValueError(
                "MergeRegImage number of transforms does not match number of images"
            )

        def get_transform_data(transform_data):
            if transform_data is not None:
                (
                    itk_composite,
                    itk_transforms,
                    final_transform,
                ) = prepare_wsireg_transform_data(transform_data)
            else:
                itk_composite, itk_transforms, final_transform = (
                    None,
                    None,
                    None,
                )

            return itk_composite, itk_transforms, final_transform

        def prepare_channel_names(sub_image_name, channel_names):
            return [f"{sub_image_name} - {c}" for c in channel_names]

        self.reg_image.channel_names = [
            prepare_channel_names(im_name, cnames)
            for im_name, cnames in zip(
                sub_image_names, self.reg_image.channel_names
            )
        ]
        self.reg_image.channel_names = [
            item
            for sublist in self.reg_image.channel_names
            for item in sublist
        ]

        merge_transform_data = [get_transform_data(t) for t in transformations]
        composite_transforms = [
            merge_transform_data[t][0]
            for t in range(len(merge_transform_data))
        ]
        final_transforms = [
            merge_transform_data[t][2]
            for t in range(len(merge_transform_data))
        ]

        spacing_check = [
            t.output_spacing for t in final_transforms if t is not None
        ]
        size_check = [t.output_size for t in final_transforms if t is not None]
        no_transform_check = [
            idx for idx, t in enumerate(final_transforms) if t is None
        ]

        for no_tform in no_transform_check:
            size_check.extend(
                [
                    np.asarray(self.reg_image.images[no_tform].im_dims)[
                        1:
                    ].tolist()
                ]
            )
            spacing_check.extend(
                [
                    [
                        self.reg_image.images[no_tform].image_res,
                        self.reg_image.images[no_tform].image_res,
                    ]
                ]
            )

        if all(spacing_check) is False:
            raise ValueError(
                "MergeRegImage all transforms output spacings and untransformed image spacings must match"
            )

        if all(size_check) is False:
            raise ValueError(
                "MergeRegImage all transforms output sizes and untransformed image sizes must match"
            )

        output_file_name = str(Path(output_dir) / f"{image_name}.ome.tiff")

        self.prepare_image_info(
            image_name,
            final_transform=final_transforms,
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
        )

        print(f"saving to {output_file_name}")
        with TiffWriter(output_file_name, bigtiff=True) as tif:
            for m_idx, merge_image in enumerate(self.reg_image.images):
                if self.reg_image.images[m_idx].reader == "sitk":
                    full_image = sitk.ReadImage(
                        self.reg_image.images[m_idx].image_filepath
                    )
                merge_n_ch = merge_image.n_ch
                for channel_idx in range(merge_n_ch):
                    if self.reg_image.images[m_idx].reader != "sitk":
                        image = self.reg_image.images[
                            m_idx
                        ].read_single_channel(channel_idx)
                        image = np.squeeze(image)
                        image = sitk.GetImageFromArray(image)
                        image.SetSpacing(
                            (
                                self.reg_image.images[m_idx].image_res,
                                self.reg_image.images[m_idx].image_res,
                            )
                        )
                    else:
                        if len(full_image.GetSize()) > 2:
                            image = full_image[:, :, channel_idx]
                        else:
                            image = full_image

                    if composite_transforms[m_idx] is not None:
                        image = transform_plane(
                            image,
                            final_transforms[m_idx],
                            composite_transforms[m_idx],
                        )

                    if isinstance(image, sitk.Image):
                        image = sitk.GetArrayFromImage(image)

                    options = dict(
                        tile=(tile_size, tile_size),
                        compression=self.compression,
                        photometric="rgb"
                        if merge_image.is_rgb
                        else "minisblack",
                        metadata=None,
                    )
                    # write OME-XML to the ImageDescription tag of the first page
                    description = (
                        self.omexml
                        if channel_idx == 0 and m_idx == 0
                        else None
                    )
                    # write channel data
                    print(
                        f" writing subimage index {m_idx} : {sub_image_names[m_idx]} - "
                        f"channel {channel_idx} - shape: {image.shape}"
                    )
                    tif.write(
                        image,
                        subifds=self.subifds,
                        description=description,
                        **options,
                    )

                    if write_pyramid:
                        for pyr_idx in range(1, self.n_pyr_levels):
                            resize_shape = (
                                self.pyr_levels[pyr_idx][0],
                                self.pyr_levels[pyr_idx][1],
                            )
                            image = cv2.resize(
                                image,
                                resize_shape,
                                cv2.INTER_LINEAR,
                            )

                            tif.write(image, **options, subfiletype=1)

            return output_file_name



import time
import yaml
from pathlib import Path
from copy import deepcopy
import json
import SimpleITK as sitk
from wsireg.reg_images.loader import reg_image_loader
from wsireg.reg_images import MergeRegImage
from wsireg.writers.merge_ome_tiff_writer import OmeTiffWriter
from wsireg.writers.merge_ome_tiff_writer import MergeOmeTiffWriter
from wsireg.utils.reg_utils import (
    register_2d_images_itkelx,
    sitk_pmap_to_dict,
    pmap_dict_to_json,
    json_to_pmap_dict,
)
from wsireg.reg_shapes import RegShapes
from wsireg.reg_transform import RegTransform
from wsireg.utils.config_utils import parse_check_reg_config
from wsireg.utils.tform_conversion import get_elastix_transforms
from wsireg.utils.tform_utils import (
    prepare_wsireg_transform_data,
)
from wsireg.utils.im_utils import (
    ARRAYLIKE_CLASSES,
)


class WsiReg2D(object):
    """
    Class to define a 2D registration graph and execute the registrations and transformations of the graph

    Parameters
    ----------
    project_name: str
        Project name will prefix all output files and directories
    output_dir: str
        Directory where registration data will be stored
    cache_images: bool
        whether to store images as they are preprocessed for registration (if you need to repeat or modify settings
        this will avoid image io and preprocessing)

    Attributes
    ----------
    modalities: dict
        dictionary of modality information (file path, spatial res., preprocessing), defines a graph node

    modalities: list
        list of all modality names

    n_modalities: int
        number of modalities (nodes) in the graphs

    reg_paths: dict
        dictionary of a modalities path from node to node

    reg_graph_edges: dict
        generated dictionary of necessary registrations to move modalities to their target modality

    n_registrations: int
        number of explicit registrations (edges) in the graphs

    transformation_paths: dict
        generated dictionary of necessary source - target transformations to transform modalities to their target modality

    transformations: dict
        per modality dictionary containing transformation parameters for each registration

    attachment_images: dict
        images to be transformed along the path of the defined graph, assoicated to a given modality (masks, other registered images)

    attachment_shapes: dict
        shape data attached to a modality to be transformed along the graph


    """

    def __init__(
        self,
        project_name: str,
        output_dir: str,
        cache_images=True,
        config=None,
    ):

        if project_name is None:
            self.project_name = 'RegProj'
        else:
            self.project_name = project_name

        if output_dir is None:
            output_dir = "./"
        self.output_dir = Path(output_dir)
        self.image_cache = self.output_dir / ".imcache_{}".format(project_name)
        self.cache_images = cache_images

        self.pairwise = False

        self._modalities = {}
        self._modality_names = []
        self._reg_paths = {}
        self._reg_graph_edges = []
        self._transform_paths = {}

        self._transformations = None

        self.n_modalities = None
        self.n_registrations = None

        self.attachment_images = {}

        self._shape_sets = {}
        self._shape_set_names = []

        self.merge_modalities = {}
        self.original_size_transforms = {}

        if config is not None:
            self.add_data_from_config(config)

    @property
    def modalities(self):
        return self._modalities

    @modalities.setter
    def modalities(self, modality):
        self._modalities.update(modality)
        self.n_modalities = len(self._modalities)

    @property
    def shape_sets(self):
        return self._shape_sets

    @shape_sets.setter
    def shape_sets(self, shape_set):
        self._shape_sets.update(shape_set)

    @property
    def shape_set_names(self):
        return self._shape_set_names

    @shape_set_names.setter
    def shape_set_names(self, shape_set_name):
        self._shape_set_names.append(shape_set_name)

    @property
    def modality_names(self):
        return self._modality_names

    @modality_names.setter
    def modality_names(self, modality_name):
        self._modality_names.append(modality_name)

    def add_modality(
        self,
        modality_name,
        image_fp,
        image_res=1,
        channel_names=None,
        channel_colors=None,
        prepro_dict={},
        mask=None,
    ):
        """
        Add an image modality (node) to the registration graph

        Parameters
        ----------
        modality_name : str
            Unique name identifier for the modality
        image_fp : str
            file path to the image to be read
        image_res : float
            spatial resolution of image in units per px (i.e. 0.9 um / px)
        prepro_dict :
            preprocessing parameters for the modality for registration. Registration images should be a xy single plane
            so many modalities (multi-channel, RGB) must "create" a single channel.
            Defaults: multi-channel images -> max intensity project image
            RGB -> greyscale then intensity inversion (black background, white foreground)
        """
        if modality_name in self._modality_names:
            raise ValueError(
                'modality named \"{}\" is already in modality_names'.format(
                    modality_name
                )
            )

        self.modalities = {
            modality_name: {
                "image_filepath": image_fp,
                "image_res": image_res,
                "channel_names": channel_names,
                "channel_colors": channel_colors,
                "preprocessing": prepro_dict,
                "mask": mask,
            }
        }

        self.modality_names = modality_name

    def add_shape_set(
        self, attachment_modality, shape_set_name, shape_files, image_res
    ):
        """
        Add a shape set to the graph

        Parameters
        ----------
        shape_set_name : str
            Unique name identifier for the shape set
        shape_files : list
            list of shape data in geoJSON format or list of dicts containing following keys:
            "array" = np.ndarray of xy coordinates, "shape_type" = geojson shape type (Polygon, MultiPolygon, etc.),
            "shape_name" = class of the shape("tumor","normal",etc.)
        image_res : float
            spatial resolution of shape data's associated image in units per px (i.e. 0.9 um / px)
        attachment_modality :
            image modality to which the shapes are attached
        """
        if shape_set_name in self._shape_set_names:
            raise ValueError(
                'shape named \"{}\" is already in shape_set_names'.format(
                    shape_set_name
                )
            )

        self.shape_sets = {
            shape_set_name: {
                "shape_files": shape_files,
                "image_res": image_res,
                "attachment_modality": attachment_modality,
            }
        }

        self.shape_set_names = shape_set_name

    def add_attachment_images(
        self,
        attachment_modality,
        modality_name,
        image_fp,
        image_res=1,
        channel_names=None,
        channel_colors=None,
    ):
        """
        Images which are unregistered between modalities, but are transformed following the path of one of the graph's
        modalities.

        Parameters
        ----------
        attachment_modality : str
            image modality to which the new image are attached
        modality_name : str
            name of the added attachment image
        image_fp : str
            path to the attachment modality, it will be imported and transformed without preprocessing
        image_res : float
            spatial resolution of attachment image data's in units per px (i.e. 0.9 um / px)
        """
        if attachment_modality not in self.modality_names:
            raise ValueError(
                'attachment modality named \"{}\" not found in modality_names'.format(
                    attachment_modality
                )
            )
        self.add_modality(
            modality_name,
            image_fp,
            image_res,
            channel_names=channel_names,
            channel_colors=channel_colors,
        )
        self.attachment_images[modality_name] = attachment_modality

    def add_attachment_shapes(
        self, attachment_modality, shape_set_name, shape_files
    ):
        if attachment_modality not in self.modality_names:
            raise ValueError(
                'attachment modality \"{}\" for shapes \"{}\" not found in modality_names {}'.format(
                    attachment_modality, shape_set_name, self.modality_names
                )
            )

        image_res = self.modalities[attachment_modality]["image_res"]
        self.add_shape_set(
            attachment_modality, shape_set_name, shape_files, image_res
        )

    @property
    def reg_paths(self):
        return self._reg_paths

    @reg_paths.setter
    def reg_paths(self, path_values):
        (
            src_modality,
            tgt_modality,
            thru_modality,
            reg_params,
            override_prepro,
        ) = path_values

        if thru_modality != tgt_modality:
            self._reg_paths.update(
                {src_modality: [thru_modality, tgt_modality]}
            )
        else:
            self._reg_paths.update({src_modality: [tgt_modality]})

        self.reg_graph_edges = {
            'modalities': {'source': src_modality, 'target': thru_modality},
            'params': reg_params,
            "override_prepro": override_prepro,
        }
        self.transform_paths = self._reg_paths

    def add_reg_path(
        self,
        src_modality_name,
        tgt_modality_name,
        thru_modality=None,
        reg_params=[],
        override_prepro={"source": None, "target": None},
    ):
        """
        Add registration path between modalities as well as a thru modality that describes where to attach edges.

        Parameters
        ----------
        src_modality_name : str
            modality that has been added to graph to be transformed to tgt_modality
        tgt_modality_name : str
            modality that has been added to graph that is being aligned to
        thru_modality: str
            modality that has been added to graph by which another should be run through
        reg_params:
            SimpleElastix registration parameters
        override_prepro:
            set specific preprocessing for a given registration edge for the source or target image that will override
            the set modality preprocessing FOR THIS REGISTRATION ONLY.
        """
        if src_modality_name not in self.modality_names:
            raise ValueError("source modality not found!")
        if tgt_modality_name not in self.modality_names:
            raise ValueError("target modality not found!")

        if thru_modality is None:
            self.reg_paths = (
                src_modality_name,
                tgt_modality_name,
                tgt_modality_name,
                reg_params,
                override_prepro,
            )
        else:
            self.reg_paths = (
                src_modality_name,
                tgt_modality_name,
                thru_modality,
                reg_params,
                override_prepro,
            )

    @property
    def reg_graph_edges(self):
        return self._reg_graph_edges

    @reg_graph_edges.setter
    def reg_graph_edges(self, edge):
        self._reg_graph_edges.append(edge)
        self.n_registrations = len(self._reg_graph_edges)

    @property
    def transform_paths(self):
        return self._transform_paths

    @transform_paths.setter
    def transform_paths(self, reg_paths):

        transform_path_dict = {}

        for k, v in reg_paths.items():
            tform_path = self.find_path(k, v[-1])
            if self.pairwise is True:
                tform_path_modalities = tform_path[:1]
            else:
                tform_path_modalities = tform_path
            transform_edges = []
            for modality in tform_path_modalities:
                for edge in self.reg_graph_edges:
                    edge_modality = edge["modalities"]['source']
                    if modality == edge_modality:
                        transform_edges.append(edge["modalities"])
            transform_path_dict.update({k: transform_edges})

        self._transform_paths = transform_path_dict

    def find_path(self, start_modality, end_modality, path=None):
        """
        Find a path from start_modality to end_modality in the graph
        """
        if path is None:
            path = []
        path = path + [start_modality]
        if start_modality == end_modality:
            return path
        if start_modality not in self.reg_paths:
            return None
        for modality in self.reg_paths[start_modality]:
            if modality not in path:
                extended_path = self.find_path(modality, end_modality, path)
                if extended_path:
                    return extended_path
        return None

    def _check_cache_modality(self, modality_name):
        cache_im_fp = self.image_cache / "{}_prepro.tiff".format(modality_name)
        cache_transform_fp = cache_im_fp.parent / "{}_init_tforms.json".format(
            cache_im_fp.stem
        )
        cache_osize_tform_fp = (
            self.image_cache
            / "{}_orig_size_tform.json".format(cache_im_fp.stem)
        )

        if cache_im_fp.exists() is True:
            im_fp = str(cache_im_fp)
            im_from_cache = True
        else:
            im_fp = self.modalities[modality_name]["image_filepath"]
            im_from_cache = False

        if cache_transform_fp.exists() is True:
            im_initial_transforms = [json_to_pmap_dict(cache_transform_fp)]
        else:
            im_initial_transforms = None

        if cache_osize_tform_fp.exists() is True:
            osize_tform = json_to_pmap_dict(cache_osize_tform_fp)
        else:
            osize_tform = None

        return im_fp, im_initial_transforms, im_from_cache, osize_tform

    def _prepare_modality(self, modality_name, reg_edge, src_or_tgt):
        mod_data = self.modalities[modality_name].copy()

        if reg_edge.get("override_prepro") is not None:
            override_preprocessing = reg_edge.get("override_prepro")[
                src_or_tgt
            ]
        else:
            override_preprocessing = None

        if override_preprocessing is not None:
            mod_data["preprocessing"] = override_preprocessing

            return (
                mod_data["image_filepath"],
                mod_data["image_res"],
                mod_data["preprocessing"],
                None,
                mod_data["mask"],
                None,
            )
        else:

            (
                mod_data["image_filepath"],
                mod_data["transforms"],
                im_from_cache,
                original_size_transform,
            ) = self._check_cache_modality(modality_name)

            if im_from_cache is True:
                if mod_data["preprocessing"].get("use_mask") is False:
                    mod_data["mask"] = None
                mod_data["preprocessing"] = None

            return (
                mod_data["image_filepath"],
                mod_data["image_res"],
                mod_data["preprocessing"],
                mod_data["transforms"],
                mod_data["mask"],
                original_size_transform,
            )

    def _cache_images(self, modality_name, reg_image):

        cache_im_fp = self.image_cache / "{}_prepro.tiff".format(modality_name)

        cache_transform_fp = self.image_cache / "{}_init_tforms.json".format(
            cache_im_fp.stem
        )

        cache_osize_tform_fp = (
            self.image_cache
            / "{}_orig_size_tform.json".format(cache_im_fp.stem)
        )
        if cache_im_fp.is_file() is False:
            sitk.WriteImage(reg_image.reg_image, str(cache_im_fp), True)

        if reg_image.mask is not None:
            cache_mask_im_fp = self.image_cache / "{}_prepro_mask.tiff".format(
                modality_name
            )
            if cache_mask_im_fp.is_file() is False:
                sitk.WriteImage(reg_image.mask, str(cache_mask_im_fp), True)

        if cache_transform_fp.is_file() is False:
            pmap_dict_to_json(
                reg_image.pre_reg_transforms, str(cache_transform_fp)
            )

        if (
            cache_osize_tform_fp.is_file() is False
            and reg_image.original_size_transform is not None
        ):
            pmap_dict_to_json(
                reg_image.original_size_transform, str(cache_osize_tform_fp)
            )

    def _find_nonreg_modalities(self):
        registered_modalities = [
            edge.get("modalities").get("source")
            for edge in self.reg_graph_edges
        ]
        non_reg_modalities = list(
            set(self.modality_names).difference(registered_modalities)
        )

        # remove attachment modalities
        for attachment_modality in self.attachment_images.keys():
            non_reg_modalities.pop(
                non_reg_modalities.index(attachment_modality)
            )

        return non_reg_modalities

    def save_config(self, registered=False):
        ts = time.strftime('%Y%m%d-%H%M%S')
        status = "registered" if registered is True else "setup"

        reg_paths = {}
        for idx, edge in enumerate(self.reg_graph_edges):
            src_modality = edge.get("modalities").get("source")
            if len(self.reg_paths[src_modality]) > 1:
                thru_modality = self.reg_paths[src_modality][0]
            else:
                thru_modality = None
            tgt_modality = self.reg_paths[src_modality][-1]
            reg_paths.update(
                {
                    f"reg_path_{idx}": {
                        "src_modality_name": edge.get("modalities").get(
                            "source"
                        ),
                        "tgt_modality_name": tgt_modality,
                        "thru_modality": thru_modality,
                        "reg_params": edge.get("params"),
                    }
                }
            )

        reg_graph_edges = deepcopy(self.reg_graph_edges)

        [rge.pop("reg_transforms", None) for rge in reg_graph_edges]

        modalities_out = deepcopy(self.modalities)
        for mod, data in modalities_out.items():
            if isinstance(data["image_filepath"], ARRAYLIKE_CLASSES):
                data["image_filepath"] = "ArrayLike"

        config = {
            "project_name": self.project_name,
            "output_dir": str(self.output_dir),
            "cache_images": self.cache_images,
            "modalities": modalities_out,
            "reg_paths": reg_paths,
            "reg_graph_edges": reg_graph_edges
            if status == "registered"
            else None,
            "original_size_transforms": self.original_size_transforms
            if status == "registered"
            else None,
            "attachment_shapes": self.shape_sets
            if len(self._shape_sets) > 0
            else None,
            "attachment_images": self.attachment_images
            if len(self.attachment_images) > 0
            else None,
        }

        output_path = (
            self.output_dir
            / f"{ts}-{self.project_name}-configuration-{status}.yaml"
        )

        with open(str(output_path), "w") as f:
            yaml.dump(config, f, sort_keys=False)

    def register_images(self, parallel=False):
        """
        Start image registration process for all modalities

        Parameters
        ----------
        parallel : bool
            whether to run each edge in parallel (not implemented yet)
        """
        if self.cache_images is True:
            self.image_cache.mkdir(parents=False, exist_ok=True)

        self.save_config(registered=False)

        for reg_edge in self.reg_graph_edges:
            if (
                reg_edge.get("registered") is None
                or reg_edge.get("registered") is False
            ):
                src_name = reg_edge["modalities"]["source"]
                tgt_name = reg_edge["modalities"]["target"]

                (
                    src_reg_image_fp,
                    src_res,
                    src_prepro,
                    src_transforms,
                    src_mask,
                    src_original_size_transform,
                ) = self._prepare_modality(src_name, reg_edge, "source")

                (
                    tgt_reg_image_fp,
                    tgt_res,
                    tgt_prepro,
                    tgt_transforms,
                    tgt_mask,
                    tgt_original_size_transform,
                ) = self._prepare_modality(tgt_name, reg_edge, "target")

                src_reg_image = reg_image_loader(
                    src_reg_image_fp,
                    src_res,
                    preprocessing=src_prepro,
                    pre_reg_transforms=src_transforms,
                    mask=src_mask,
                )

                tgt_reg_image = reg_image_loader(
                    tgt_reg_image_fp,
                    tgt_res,
                    preprocessing=tgt_prepro,
                    pre_reg_transforms=tgt_transforms,
                    mask=tgt_mask,
                )

                src_reg_image.read_reg_image()
                tgt_reg_image.read_reg_image()

                if (
                    tgt_original_size_transform is None
                    and tgt_reg_image.original_size_transform is not None
                ):
                    tgt_original_size_transform = (
                        tgt_reg_image.original_size_transform
                    )

                if self.cache_images is True:
                    if reg_edge.get("override_prepro") is not None:
                        if (
                            reg_edge.get("override_prepro").get("source")
                            is None
                        ):
                            self._cache_images(src_name, src_reg_image)
                        if (
                            reg_edge.get("override_prepro").get("target")
                            is None
                        ):
                            self._cache_images(tgt_name, tgt_reg_image)
                    else:
                        self._cache_images(src_name, src_reg_image)
                        self._cache_images(tgt_name, tgt_reg_image)

                reg_params = reg_edge["params"]

                output_path = (
                    self.output_dir
                    / "{}-{}_to_{}_reg_output".format(
                        self.project_name,
                        reg_edge["modalities"]["source"],
                        reg_edge["modalities"]["target"],
                    )
                )

                output_path.mkdir(parents=False, exist_ok=True)

                output_path_tform = (
                    self.output_dir
                    / "{}-{}_to_{}_transformations.json".format(
                        self.project_name,
                        reg_edge["modalities"]["source"],
                        reg_edge["modalities"]["target"],
                    )
                )

                reg_tforms = register_2d_images_itkelx(
                    src_reg_image,
                    tgt_reg_image,
                    reg_params,
                    output_path,
                )

                reg_tforms = [sitk_pmap_to_dict(tf) for tf in reg_tforms]

                if src_transforms is not None:
                    initial_transforms = src_transforms[0]
                else:
                    initial_transforms = src_reg_image.pre_reg_transforms

                reg_edge["transforms"] = {
                    'initial': initial_transforms,
                    'registration': reg_tforms,
                }

                self.original_size_transforms.update(
                    {tgt_name: tgt_original_size_transform}
                )

                reg_edge["registered"] = True
                pmap_dict_to_json(
                    reg_edge["transforms"], str(output_path_tform)
                )

        self.transformations = self.reg_graph_edges
        self.save_config(registered=True)

    @property
    def transformations(self):
        return self._transformations

    @transformations.setter
    def transformations(self, reg_graph_edges):
        self._transformations = self._collate_transformations()

    def add_merge_modalities(self, merge_name, modalities):
        for modality in modalities:
            try:
                self.modalities[modality]
            except KeyError:
                raise ValueError(
                    f"Modality for merger [{modality}] is not a modality "
                    f"within the graph, current modalitles : "
                    f"{self.modality_names}"
                )
            self.merge_modalities.update({merge_name: modalities})

    def _generate_reg_transforms(self):
        self._reg_graph_edges["reg_transforms"]

    def _collate_transformations(self):
        transforms = {}
        for reg_edge in self.reg_graph_edges:
            if reg_edge["transforms"]["initial"] is not None:
                initial_transforms = [
                    RegTransform(t) for t in reg_edge["transforms"]["initial"]
                ]
            else:
                initial_transforms = []
            reg_edge["reg_transforms"] = {
                'initial': initial_transforms,
                'registration': [
                    RegTransform(t)
                    for t in reg_edge["transforms"]["registration"]
                ],
            }
        edge_modality_pairs = [v['modalities'] for v in self.reg_graph_edges]
        for modality, tform_edges in self.transform_paths.items():
            for idx, tform_edge in enumerate(tform_edges):
                reg_edge_tforms = self.reg_graph_edges[
                    edge_modality_pairs.index(tform_edge)
                ]["reg_transforms"]
                if idx == 0:
                    transforms[modality] = {
                        'initial': reg_edge_tforms['initial'],
                        idx: reg_edge_tforms['registration'],
                    }
                else:
                    transforms[modality][idx] = reg_edge_tforms['registration']

        return transforms

    def _prepare_nonreg_image_transform(
        self, modality_key, to_original_size=True
    ):
        print(
            "transforming non-registered modality : {} ".format(modality_key)
        )
        output_path = self.output_dir / "{}-{}_registered".format(
            self.project_name, modality_key
        )
        im_data = self.modalities[modality_key]
        transformations = {"initial": None, "registered": None}

        if (
            im_data.get("preprocessing").get("rot_cc") is not None
            or im_data.get("preprocessing").get("flip") is not None
            or im_data.get("preprocessing").get("mask_to_bbox") is True
            or im_data.get("preprocessing").get("mask_bbox") is not None
        ):

            transformations.update(
                {
                    "initial": self._check_cache_modality(modality_key)[1][0],
                    "registered": None,
                }
            )

        if to_original_size is True:
            transformations.update(
                {"registered": self.original_size_transforms.get(modality_key)}
            )

        if (
            transformations.get("initial") is None
            and transformations.get("registered") is None
        ):
            transformations = None

        return im_data, transformations, output_path

    def _prepare_reg_image_transform(
        self,
        edge_key,
        attachment=False,
        attachment_modality=None,
        to_original_size=True,
    ):
        im_data = self.modalities[edge_key]

        if attachment is True:
            final_modality = self.reg_paths[attachment_modality][-1]
            transformations = deepcopy(
                self.transformations[attachment_modality]
            )
        else:
            final_modality = self.reg_paths[edge_key][-1]
            transformations = deepcopy(self.transformations[edge_key])

        print("transforming {} to {}".format(edge_key, final_modality))

        output_path = self.output_dir / "{}-{}_to_{}_registered".format(
            self.project_name,
            edge_key,
            final_modality,
        )
        if (
            self.original_size_transforms.get(final_modality) is not None
            and to_original_size is True
        ):
            original_size_transform = self.original_size_transforms[
                final_modality
            ]
            transformations.update(
                {"orig": [RegTransform(original_size_transform)]}
            )

        return im_data, transformations, output_path

    def _transform_write_image(
        self, im_data, transformations, output_path, file_writer="ome.tiff"
    ):

        tfregimage = reg_image_loader(
            im_data["image_filepath"],
            im_data["image_res"],
            channel_names=im_data.get("channel_names"),
            channel_colors=im_data.get("channel_colors"),
        )

        ometiffwriter = OmeTiffWriter(tfregimage)

        if transformations:
            (
                composite_transform,
                itk_transforms,
                final_transform,
            ) = prepare_wsireg_transform_data(transformations)
        else:
            composite_transform, itk_transforms, final_transform = (
                None,
                None,
                None,
            )

        if (
            file_writer == "ome.tiff-bytile"
            and ometiffwriter.reg_image.reader not in ["czi", "sitk"]
        ):
            im_fp = ometiffwriter.write_image_by_tile(
                output_path.stem,
                itk_transforms=itk_transforms,
                composite_transform=composite_transform,
                final_transform=final_transform,
                output_dir=str(self.output_dir),
            )
        else:
            im_fp = ometiffwriter.write_image_by_plane(
                output_path.stem,
                composite_transform=composite_transform,
                final_transform=final_transform,
                output_dir=str(self.output_dir),
            )

        return im_fp

    def _transform_write_merge_images(self, to_original_size=True):
        for merge_name, sub_images in self.merge_modalities.items():
            im_fps = []
            im_res = []
            im_ch_names = []
            transformations = []
            final_modalities = []
            for sub_image in sub_images:
                im_data = self.modalities[sub_image]
                im_fps.append(im_data["image_filepath"])
                im_res.append(im_data["image_res"])
                im_ch_names.append(im_data.get("channel_names"))

                try:
                    transforms = deepcopy(self.transformations[sub_image])
                except KeyError:
                    transforms = None

                try:
                    final_modalities.append(self.reg_paths[sub_image][-1])
                except KeyError:
                    initial_transforms = self._check_cache_modality(sub_image)[
                        1
                    ][0]
                    if initial_transforms:
                        initial_transforms = [
                            RegTransform(t) for t in initial_transforms
                        ]
                        final_modalities.append(sub_image)
                        transforms = {"initial": initial_transforms}
                    else:
                        transforms = None

                transformations.append(transforms)

            if all(final_modalities):
                final_modality = final_modalities[0]
            else:
                raise ValueError("final modalities do not match on merge")

            if (
                self.original_size_transforms.get(final_modality) is not None
                and to_original_size is True
            ):
                original_size_transform = self.original_size_transforms[
                    final_modality
                ]
                for transformation in transformations:
                    if transformation is None:
                        transformation = {}
                    transformation.update(
                        {"orig": [RegTransform(original_size_transform)]}
                    )

            output_path = self.output_dir / "{}-{}_merged-registered".format(
                self.project_name,
                merge_name,
            )

            merge_regimage = MergeRegImage(
                im_fps,
                im_res,
                channel_names=im_ch_names,
            )

            merge_ometiffwriter = MergeOmeTiffWriter(merge_regimage)

            im_fp = merge_ometiffwriter.merge_write_image_by_plane(
                output_path.stem,
                sub_images,
                transformations=transformations,
                output_dir=str(self.output_dir),
            )
            return im_fp

    def transform_images(
        self,
        file_writer="ome.tiff",
        transform_non_reg=True,
        remove_merged=True,
        to_original_size=True,
    ):
        """
        Transform and write images to disk after registration. Also transforms all attachment images

        Parameters
        ----------
        file_writer : str
            output type to use, sitk writes a single resolution tiff, "zarr" writes an ome-zarr multiscale
            zarr store
        transform_non_reg : bool
            whether to write the images that aren't transformed during registration as well
        remove_merged: bool
            whether to remove images that are stored in merged store, if True, images that are merged
            will not be written as individual images as well
        to_original_size: bool
            write images that have been cropped for registration back to their original coordinate space
        """
        image_fps = []

        if all(
            [reg_edge.get("registered") for reg_edge in self.reg_graph_edges]
        ):
            # prepare workflow
            merge_modalities = []
            if len(self.merge_modalities.keys()) > 0:
                for k, v in self.merge_modalities.items():
                    merge_modalities.extend(v)

            reg_path_keys = list(self.reg_paths.keys())
            nonreg_keys = self._find_nonreg_modalities()

            if remove_merged:
                for merge_mod in merge_modalities:
                    try:
                        m_idx = reg_path_keys.index(merge_mod)
                        reg_path_keys.pop(m_idx)
                    except ValueError:
                        pass
                    try:
                        m_idx = nonreg_keys.index(merge_mod)
                        nonreg_keys.pop(m_idx)
                    except ValueError:
                        pass

            for modality in reg_path_keys:
                (
                    im_data,
                    transformations,
                    output_path,
                ) = self._prepare_reg_image_transform(
                    modality,
                    attachment=False,
                    to_original_size=to_original_size,
                )
                im_fp = self._transform_write_image(
                    im_data,
                    transformations,
                    output_path,
                    file_writer=file_writer,
                )
                image_fps.append(im_fp)

            for (
                modality,
                attachment_modality,
            ) in self.attachment_images.items():
                (
                    im_data,
                    transformations,
                    output_path,
                ) = self._prepare_reg_image_transform(
                    modality,
                    attachment=True,
                    attachment_modality=attachment_modality,
                    to_original_size=to_original_size,
                )

                im_fp = self._transform_write_image(
                    im_data,
                    transformations,
                    output_path,
                    file_writer=file_writer,
                )
                image_fps.append(im_fp)
            if len(self.merge_modalities.items()) > 0:
                im_fp = self._transform_write_merge_images(
                    to_original_size=to_original_size
                )
                image_fps.append(im_fp)

        if transform_non_reg is True:
            # preprocess and save unregistered nodes
            for modality in nonreg_keys:
                (
                    im_data,
                    transformations,
                    output_path,
                ) = self._prepare_nonreg_image_transform(
                    modality,
                    to_original_size=to_original_size,
                )
                im_fp = self._transform_write_image(
                    im_data,
                    transformations,
                    output_path,
                    file_writer=file_writer,
                )
                image_fps.append(im_fp)

        return image_fps

    def transform_shapes(self):
        """
        Transform all attached shapes and write out shape data to geoJSON.
        """
        for set_name, set_data in self.shape_sets.items():
            attachment_modality = set_data["attachment_modality"]

            final_modality = self.reg_paths[attachment_modality][-1]

            print(
                "transforming shape set {} associated with {} to {}".format(
                    set_name, attachment_modality, final_modality
                )
            )

            rs = RegShapes(
                set_data["shape_files"], source_res=set_data["image_res"]
            )
            rs.transform_shapes(
                self.transformations[attachment_modality],
            )

            output_path = (
                self.output_dir
                / "{}-{}-{}_to_{}-transformed_shapes.geojson".format(
                    self.project_name,
                    set_name,
                    attachment_modality,
                    final_modality,
                )
            )

            rs.save_shape_data(output_path, transformed=True)

    def save_transformations(self):
        """
        Save all transformations for a given modality as JSON
        """
        if all(
            [reg_edge.get("registered") for reg_edge in self.reg_graph_edges]
        ):
            for key in self.reg_paths.keys():

                final_modality = self.reg_paths[key][-1]

                output_path = (
                    self.output_dir
                    / "{}-{}_to_{}_transformations.json".format(
                        self.project_name,
                        key,
                        final_modality,
                    )
                )
                out_transforms = get_elastix_transforms(
                    self.transformations[key]
                )

                with open(output_path, 'w') as fp:
                    json.dump(out_transforms, fp, indent=4)

            for (
                modality,
                attachment_modality,
            ) in self.attachment_images.items():

                final_modality = self.reg_paths[attachment_modality][-1]

                output_path = (
                    self.output_dir
                    / "{}-{}_to_{}_transformations.json".format(
                        self.project_name,
                        modality,
                        final_modality,
                    )
                )

                out_transforms = get_elastix_transforms(
                    self.transformations[attachment_modality]
                )

                with open(output_path, 'w') as fp:
                    json.dump(out_transforms, fp, indent=4)
        else:
            print("registration has not been executed for the graph")

    def add_data_from_config(self, config_filepath):

        reg_config = parse_check_reg_config(config_filepath)

        if reg_config.get("modalities") is not None:
            for key, val in reg_config["modalities"].items():

                image_filepath = (
                    val.get("image_filepath")
                    if val.get("image_filepath") is not None
                    else val.get("image_filepath")
                )

                preprocessing = (
                    "None"
                    if val.get("preprocessing") is None
                    else val.get("preprocessing")
                )

                self.add_modality(
                    key,
                    image_filepath,
                    image_res=val.get("image_res"),
                    channel_names=val.get("channel_names"),
                    channel_colors=val.get("channel_colors"),
                    prepro_dict=preprocessing,
                    mask=val.get("mask"),
                )
        else:
            print("warning: config file did not contain any image modalities")

        if reg_config.get("reg_paths") is not None:

            for key, val in reg_config["reg_paths"].items():
                self.add_reg_path(
                    val.get("src_modality_name"),
                    val.get("tgt_modality_name"),
                    val.get("thru_modality"),
                    reg_params=val.get("reg_params"),
                    override_prepro=val.get("override_prepro"),
                )
        else:
            print(
                "warning: config file did not contain any registration paths"
            )
        if reg_config.get("attachment_images") is not None:

            for key, val in reg_config["attachment_images"].items():
                self.add_attachment_images(
                    val.get("attachment_modality"),
                    key,
                    val.get("image_filepath"),
                    val.get("image_res"),
                    channel_names=val.get("channel_names"),
                    channel_colors=val.get("channel_colors"),
                )

        if reg_config.get("attachment_shapes") is not None:

            for key, val in reg_config["attachment_shapes"].items():
                self.add_attachment_shapes(
                    val.get("attachment_modality"), key, val.get("shape_files")
                )

        if reg_config.get("reg_graph_edges") is not None:
            self._reg_graph_edges = reg_config["reg_graph_edges"]
            if all([re.get("registered") for re in self.reg_graph_edges]):
                self.transformations = self.reg_graph_edges

        if reg_config.get("merge_modalities") is not None:
            for mn, mm in reg_config["merge_modalities"]:
                self.add_merge_modalities(mn, mm)

    def reset_registered_modality(self, modalities):
        edge_keys = [
            r.get("modalities").get("source") for r in self.reg_graph_edges
        ]
        if isinstance(modalities, str):
            modalities = [modalities]

        for modality in modalities:
            modality_idx = edge_keys.index(modality)
            self.reg_graph_edges[modality_idx]["registered"] = False


def main():
    import argparse

    def config_to_WsiReg2D(config_filepath):
        reg_config = parse_check_reg_config(config_filepath)

        reg_graph = WsiReg2D(
            reg_config.get("project_name"),
            reg_config.get("output_dir"),
            reg_config.get("cache_images"),
        )
        return reg_graph

    parser = argparse.ArgumentParser(
        description='Load Whole Slide Image 2D Registration Graph from configuration file'
    )

    parser.add_argument(
        "config_filepath",
        metavar="C",
        type=str,
        nargs=1,
        help="full filepath for .yaml configuration file",
    )
    parser.add_argument(
        "--fw",
        type=str,
        nargs=1,
        help="how to write output registered images: ome.tiff, ome.zarr (default: ome.tiff)",
    )

    args = parser.parse_args()
    config_filepath = args.config_filepath[0]
    if args.fw is None:
        file_writer = "ome.tiff"
    else:
        file_writer = args.fw[0]

    reg_graph = config_to_WsiReg2D(config_filepath)
    reg_graph.add_data_from_config(config_filepath)

    reg_graph.register_images()
    reg_graph.save_transformations()
    reg_graph.transform_images(file_writer=file_writer)

    if reg_graph.shape_sets:
        reg_graph.transform_shapes()


if __name__ == "__main__":
    import sys

    sys.exit(main())



import warnings
import SimpleITK as sitk
from tifffile import TiffFile, xml2dict
from wsireg.reg_images.reg_image import RegImage
from wsireg.utils.im_utils import (
    std_prepro,
    guess_rgb,
    get_tifffile_info,
    tf_get_largest_series,
    tifffile_dask_backend,
    tifffile_zarr_backend,
    tf_zarr_read_single_ch,
)


class OmeTiffRegImage(RegImage):
    def __init__(
        self,
        image_fp,
        image_res,
        mask=None,
        pre_reg_transforms=None,
        preprocessing=None,
        channel_names=None,
        channel_colors=None,
    ):
        self.image_filepath = image_fp
        self.image_res = image_res
        self.image = None
        self.tf = TiffFile(self.image_filepath)
        self.reader = "tifffile"

        self.ome_metadata = xml2dict(self.tf.ome_metadata)

        (
            self.im_dims,
            self.im_dtype,
        ) = self._get_image_info()

        self.im_dims = tuple(self.im_dims)

        # self.n_ch = self.im_dims[2] if self.is_rgb else self.im_dims[0]
        self.mask = self.read_mask(mask)

        if preprocessing is None:
            self.preprocessing = std_prepro()
        else:
            self.preprocessing = std_prepro()
            self.preprocessing.update(preprocessing)

        self.pre_reg_transforms = pre_reg_transforms

        self.channel_names = channel_names
        self.channel_colors = channel_colors
        self.original_size_transform = None

    def _get_image_info(self):
        if len(self.tf.series) > 1:
            warnings.warn(
                "The tiff contains multiple series, "
                "the largest series will be read by default"
            )
            largest_series = tf_get_largest_series(self.image_filepath)
            series_metadata = self.ome_metadata.get("OME").get("Image")[
                largest_series
            ]
        else:
            series_metadata = self.ome_metadata.get("OME").get("Image")

        im_dims, im_dtype = get_tifffile_info(self.image_filepath)

        if isinstance(series_metadata.get("Pixels").get("Channel"), list):
            samples_per_pixel = (
                series_metadata.get("Pixels")
                .get("Channel")[0]
                .get("SamplesPerPixel")
            )
        else:
            samples_per_pixel = (
                series_metadata.get("Pixels")
                .get("Channel")
                .get("SamplesPerPixel")
            )

        is_rgb = guess_rgb(im_dims)
        self.n_ch = im_dims[2] if is_rgb else im_dims[0]

        if is_rgb is False and samples_per_pixel >= 3:
            self.is_rgb = True
            self.is_rgb_interleaved = False
        elif is_rgb is True and samples_per_pixel >= 3:
            self.is_rgb = True
            self.is_rgb_interleaved = True
        else:
            self.is_rgb = False
            self.is_rgb_interleaved = False

        return im_dims, im_dtype

    def read_reg_image(self):
        largest_series = tf_get_largest_series(self.image_filepath)

        if self.is_rgb is True and self.is_rgb_interleaved is False:
            force_rgb = True
        else:
            force_rgb = None

        try:
            image = tifffile_dask_backend(
                self.image_filepath,
                largest_series,
                self.preprocessing,
                force_rgb=force_rgb,
            )
        except ValueError:
            image = tifffile_zarr_backend(
                self.image_filepath,
                largest_series,
                self.preprocessing,
                force_rgb=force_rgb,
            )

        if (
            self.preprocessing is not None
            and self.preprocessing.get('as_uint8') is True
            and image.GetPixelID() != sitk.sitkUInt8
        ):
            image = sitk.RescaleIntensity(image)
            image = sitk.Cast(image, sitk.sitkUInt8)

        image, spatial_preprocessing = self.preprocess_reg_image_intensity(
            image, self.preprocessing
        )

        if image.GetDepth() >= 1:
            raise ValueError(
                "preprocessing did not result in a single image plane\n"
                "multi-channel or 3D image return"
            )

        if image.GetNumberOfComponentsPerPixel() > 1:
            raise ValueError(
                "preprocessing did not result in a single image plane\n"
                "multi-component / RGB(A) image returned"
            )

        if (
            len(spatial_preprocessing) > 0
            or self.pre_reg_transforms is not None
        ):
            (
                self.image,
                self.pre_reg_transforms,
            ) = self.preprocess_reg_image_spatial(
                image, spatial_preprocessing, self.pre_reg_transforms
            )
        else:
            self.image = image
            self.pre_reg_transforms = None

    def read_single_channel(self, channel_idx: int):
        if channel_idx > (self.n_ch - 1):
            warnings.warn(
                "channel_idx exceeds number of channels, reading channel at channel_idx == 0"
            )
            channel_idx = 0
        image = tf_zarr_read_single_ch(
            self.image_filepath,
            channel_idx,
            self.is_rgb,
            is_rgb_interleaved=self.is_rgb_interleaved,
        )
        return image



import multiprocessing
from concurrent.futures import ThreadPoolExecutor
import SimpleITK as sitk
import numpy as np
import zarr
import dask.array as da
from wsireg.utils.tform_utils import (
    ELX_TO_ITK_INTERPOLATORS,
)
from wsireg.utils.shape_utils import (
    prepare_pt_transformation_data,
)


def tile_pad_output_size(y_size, x_size, tile_size=512):
    y_size, x_size = (
        np.ceil(np.asarray((y_size, x_size)) / tile_size) * tile_size
    ).astype(int)
    return y_size, x_size


def generate_tile_coords(output_size_xy, output_spacing, tile_size=512):

    image = sitk.Image(
        np.asarray(output_size_xy).astype(np.int32).tolist(), sitk.sitkUInt8
    )
    image.SetSpacing(output_spacing)
    x_size, y_size = image.GetSize()
    tile_coordinate_data = []
    for y in range(0, y_size, tile_size):
        for x in range(0, x_size, tile_size):
            tile_origin = image.TransformIndexToPhysicalPoint([x, y])
            tile_data = {
                "tile_index": (x, y),
                "tile_origin": tile_origin,
                "output_size": (tile_size, tile_size),
                "output_spacing": output_spacing,
            }
            tile_coordinate_data.append(tile_data)

    return tile_coordinate_data


def add_tile_location_on_moving(
    wsireg_transforms, tile_coordinate_data, source_res=1
):

    pt_transform, target_res = prepare_pt_transformation_data(
        wsireg_transforms, compute_inverse=False
    )

    # ordering is to transform moving to fixed
    # reversing goes from fixed to moving
    pt_transform = pt_transform[::-1]

    for tile in tile_coordinate_data:
        x0 = tile["tile_index"][0]
        y0 = tile["tile_index"][1]
        x1 = x0 + tile["output_size"][0]
        y1 = y0 + tile["output_size"][1]
        tile_corners = np.asarray([[x0, y0], [x1, y0], [x1, y1], [x0, y1]])

        tformed_pts = []
        for pt in tile_corners:
            pt = pt * target_res
            for idx, t in enumerate(pt_transform):
                if idx == 0:
                    t_pt = t.itk_transform.TransformPoint(pt)
                else:
                    t_pt = t.itk_transform.TransformPoint(t_pt)
            t_pt = np.array(t_pt)
            t_pt /= source_res
            tformed_pts.append(t_pt)

        transformed_tile_pts = np.stack(tformed_pts)
        tile.update({"tile_on_moving": transformed_tile_pts})

    return tile_coordinate_data


def pad_transform_tile_corners(
    x_size, y_size, transformed_tile_pts, padding=128
):
    x_min = np.floor(np.min(transformed_tile_pts[:, 0]))
    y_min = np.floor(np.min(transformed_tile_pts[:, 1]))
    x_max = np.ceil(np.max(transformed_tile_pts[:, 0]))
    y_max = np.ceil(np.max(transformed_tile_pts[:, 1]))

    if (x_min - padding) < 0:
        x_min = 0
    else:
        x_min -= padding

    if (y_min - padding) < 0:
        y_min = 0
    else:
        y_min -= padding

    if (x_max + padding) > x_size:
        x_max = x_size
    else:
        if (x_max + padding) > 0:
            x_max += padding
        else:
            x_max = x_size

    if (y_max + padding) > y_size:
        y_max = y_size
    else:
        if (y_max + padding) > 0:
            y_max += padding
        else:
            y_max = y_size

    return int(x_min), int(x_max), int(y_min), int(y_max)


def wsireg_tile_transform_to_resampler(
    tile_data, final_transform_interpolator="FinalNearestNeighborInterpolator"
):
    resampler = sitk.ResampleImageFilter()
    resampler.SetOutputOrigin(tile_data["tile_origin"])
    resampler.SetSize(tile_data["output_size"])
    resampler.SetOutputSpacing(tile_data["output_spacing"])
    interpolator = ELX_TO_ITK_INTERPOLATORS.get(final_transform_interpolator)
    resampler.SetInterpolator(interpolator)
    return resampler


def determine_moving_tile_padding(itk_transforms):
    transform_linearity = [t.is_linear for t in itk_transforms]
    if all(transform_linearity):
        tile_padding = 256
    else:
        nl_idxs = np.where(np.array(transform_linearity) == 0)[0]
        for nl_idx in nl_idxs:
            output_spacing = itk_transforms[nl_idx].output_spacing[0]
            grid_spacing = itk_transforms[
                nl_idx
            ].itk_transform.GetFixedParameters()[4:6]
            tile_padding = (np.max(grid_spacing) / output_spacing) * 12
    return tile_padding


def itk_transform_tiles(
    tform_reg_im,
    y_size,
    x_size,
    tile_coordinate_data,
    itk_transform_composite,
    ch_idx=0,
    tile_size=512,
    tile_padding=128,
    use_multiprocessing=False,
):

    zstr = zarr.TempStore()
    zgrp = zarr.open(zstr)
    if tform_reg_im.is_rgb:
        resample_zarray = zgrp.create_dataset(
            'temparray',
            shape=(y_size, x_size, tform_reg_im.im_dims[-1]),
            chunks=(tile_size, tile_size),
            dtype=tform_reg_im.im_dtype,
        )
    else:
        resample_zarray = zgrp.create_dataset(
            'temparray',
            shape=(y_size, x_size),
            chunks=(tile_size, tile_size),
            dtype=tform_reg_im.im_dtype,
        )
    if tform_reg_im.is_rgb:
        y_size_moving = tform_reg_im.im_dims[0]
        x_size_moving = tform_reg_im.im_dims[1]
    else:
        y_size_moving = tform_reg_im.im_dims[1]
        x_size_moving = tform_reg_im.im_dims[2]

    def transform_tile(tile):
        tile_resampler = wsireg_tile_transform_to_resampler(tile)
        x_min, x_max, y_min, y_max = pad_transform_tile_corners(
            x_size_moving,
            y_size_moving,
            tile["tile_on_moving"],
            padding=tile_padding,
        )

        x_idx_start = tile["tile_index"][0]
        x_idx_end = x_idx_start + tile["output_size"][0]
        y_idx_start = tile["tile_index"][1]
        y_idx_end = y_idx_start + tile["output_size"][1]

        if x_min == 0 and x_max == 0:
            x_max = 1
        if y_min == 0 and y_max == 0:
            y_max = 1

        if tform_reg_im.is_rgb:
            image = tform_reg_im.image[y_min:y_max, x_min:x_max, :]
            image = sitk.GetImageFromArray(image, isVector=True)
            image.SetSpacing((tform_reg_im.image_res, tform_reg_im.image_res))
            image.SetOrigin(
                image.TransformIndexToPhysicalPoint([int(x_min), int(y_min)])
            )

        elif tform_reg_im.n_ch == 1:
            image = tform_reg_im.image[y_min:y_max, x_min:x_max]
            image = sitk.GetImageFromArray(image, isVector=False)
            image.SetSpacing((tform_reg_im.image_res, tform_reg_im.image_res))
            image.SetOrigin(
                image.TransformIndexToPhysicalPoint([int(x_min), int(y_min)])
            )
        else:
            image = tform_reg_im.image[ch_idx, y_min:y_max, x_min:x_max]
            image = sitk.GetImageFromArray(image, isVector=False)
            image.SetSpacing((tform_reg_im.image_res, tform_reg_im.image_res))
            image.SetOrigin(
                image.TransformIndexToPhysicalPoint([int(x_min), int(y_min)])
            )

        tile_resampler.SetTransform(itk_transform_composite)
        tile_resampled = tile_resampler.Execute(image)

        if tform_reg_im.is_rgb:
            resample_zarray[
                y_idx_start:y_idx_end, x_idx_start:x_idx_end, :
            ] = sitk.GetArrayFromImage(tile_resampled)
        else:
            resample_zarray[
                y_idx_start:y_idx_end, x_idx_start:x_idx_end
            ] = sitk.GetArrayFromImage(tile_resampled)

    if use_multiprocessing:
        max_workers = multiprocessing.cpu_count() - 1
        with ThreadPoolExecutor(max_workers) as executor:
            executor.map(transform_tile, tile_coordinate_data)
    else:
        for tile in tile_coordinate_data:
            transform_tile(tile)

    return resample_zarray


def zarr_to_tiles(z, tile_coordinate_data, is_rgb):
    for tile in tile_coordinate_data:
        x_idx_start = tile["tile_index"][0]
        x_idx_end = x_idx_start + tile["output_size"][0]
        y_idx_start = tile["tile_index"][1]
        y_idx_end = y_idx_start + tile["output_size"][1]
        if is_rgb:
            yield z[y_idx_start:y_idx_end, x_idx_start:x_idx_end, :]
        else:
            yield z[y_idx_start:y_idx_end, x_idx_start:x_idx_end]


def subres_zarr_to_tiles(z, tile_size, is_rgb):
    for y in range(0, z.shape[0], tile_size):
        for x in range(0, z.shape[1], tile_size):
            if is_rgb:
                yield z[y : y + tile_size, x : x + tile_size, :].compute()
            else:
                yield z[y : y + tile_size, x : x + tile_size].compute()


def compute_sub_res(
    tform_reg_im, resampled_zarray, ds_factor, tile_size, is_rgb
):
    if is_rgb:
        resampling_axis = {0: ds_factor, 1: ds_factor, 2: 1}
        tiling = (tile_size, tile_size, 3)
    else:
        resampling_axis = {0: ds_factor, 1: ds_factor}
        tiling = (tile_size, tile_size)

    resampled_zarray_subres = da.coarsen(
        np.mean,
        da.from_zarr(resampled_zarray),
        resampling_axis,
        trim_excess=True,
    )
    resampled_zarray_subres = resampled_zarray_subres.astype(
        tform_reg_im.im_dtype
    )
    resampled_zarray_subres = resampled_zarray_subres.rechunk(tiling)
    orig_shape = resampled_zarray_subres.shape

    y_size_pyr, x_size_pyr = tile_pad_output_size(
        resampled_zarray_subres.shape[0],
        resampled_zarray_subres.shape[1],
        tile_size=tile_size,
    )
    y_pad = y_size_pyr - resampled_zarray_subres.shape[0]
    x_pad = x_size_pyr - resampled_zarray_subres.shape[1]
    if is_rgb:
        padding = ((0, y_pad), (0, x_pad), (0, 0))
    else:
        padding = ((0, y_pad), (0, x_pad))

    resampled_zarray_subres = da.pad(resampled_zarray_subres, padding)
    return resampled_zarray_subres, orig_shape



from copy import deepcopy
import numpy as np
import SimpleITK as sitk
import itk
from wsireg.utils.im_utils import (
    contrast_enhance,
    sitk_inv_int,
    transform_to_ome_zarr,
    compute_mask_to_bbox,
    transform_plane,
)
from wsireg.utils.tform_utils import (
    gen_aff_tform_flip,
    gen_rigid_tform_rot,
    gen_rigid_translation,
    gen_rig_to_original,
    prepare_wsireg_transform_data,
)
from wsireg.writers.ome_tiff_writer import OmeTiffWriter


class RegImage:
    """
    Image container class for reading images and preparing them for registration.
    Does two forms of preprocessing:
    Intensity preprocessing, modifying intensity values for best registration
    Spatial preprocessing, rotating and flipping images prior to registration for best alignment

    After preprocessing, images destined for registration should always be a single channel 2D image

    Parameters
    ----------
    image : str or np.array
        filepath to image to be processed or a numpy array of the image
    image_res : float
        Physical spacing (xy resolution) of the image (units are assumed to be the same for each image in experiment but are not
        defined)
        Does not yet support isotropic resolution
    prepro_dict : dict
        preprocessing dict to apply to image
    transforms:list
        list of elastix transformation data to apply to an image during preprocessing
    mask : str or np.ndarray
        filepath of a mask image to be processed or a numpy array of the mask
        all values => 1 are considered WITHIN the mask
    """

    def __init__(self):
        self.image = None
        self.reg_image = None
        self.original_size_transform = None
        return

    def read_mask(self, mask):
        if mask is not None:
            if isinstance(mask, np.ndarray):
                mask = sitk.GetImageFromArray(mask)
            elif isinstance(mask, str):
                mask = sitk.ReadImage(mask)
            elif isinstance(mask, sitk.Image):
                mask = mask

            mask.SetSpacing((self.image_res, self.image_res))

        return mask

    def preprocess_reg_image_intensity(self, image, preprocessing):
        # if isinstance(image, Path):
        #     image = read_image(str(image), preprocessing=preprocessing)
        # else:
        #     image = prepare_np_image(image, preprocessing=preprocessing)

        # separate spatial preprocessing from intensity preprocessing
        spatial_preprocessing = {}
        for spat_key in [
            "mask_to_bbox",
            "mask_bbox",
            "rot_cc",
            "flip",
            "use_mask",
        ]:
            if spat_key in preprocessing:
                spatial_preprocessing[spat_key] = preprocessing[spat_key]
                preprocessing.pop(spat_key, None)

        # remove read time preprocessing
        preprocessing.pop("ch_indices", None)
        preprocessing.pop("as_uint8", None)

        # type specific
        if preprocessing["image_type"] == "FL":
            preprocessing.pop("inv_int", None)
        elif preprocessing["image_type"] == "BF":
            preprocessing.pop("max_int_proj", None)
            preprocessing.pop("contrast_enhance", None)

        if preprocessing.get("contrast_enhance_opt") is True:
            preprocessing.update({"contrast_enhance": contrast_enhance})
        else:
            preprocessing.pop("contrast_enhance", None)

        if preprocessing.get("inv_int_opt") is True or self.is_rgb is True:
            preprocessing.update({"inv_int": sitk_inv_int})
        else:
            preprocessing.pop("inv_int", None)

        preprocessing.pop("contrast_enhance_opt", None)
        preprocessing.pop("inv_int_opt", None)

        # remove type and set downsample to last
        preprocessing.pop("image_type", None)

        spatial_preprocessing.update(
            {"downsample": preprocessing.get("downsample")}
        )
        preprocessing.pop("downsample", None)

        # iterate through intensity transformations preprocessing dictionary
        for k, v in preprocessing.items():
            if v is not None:
                print("performing preprocessing: ", k)
                image = v(image)

        image.SetSpacing((self.image_res, self.image_res))

        return image, spatial_preprocessing

    def preprocess_reg_image_spatial(
        self, image, spatial_preprocessing, imported_transforms=None
    ):
        # spatial preprocessing:
        # imported transforms -> Masking -> rotation -> flipping
        transforms = []
        # if imported_transforms is not None:
        #     image = transform_2d_image_itkelx(
        #         image, self.image_res, imported_transforms
        #     )
        #     if self.mask is not None:
        #         self.mask = transform_2d_image_itkelx(
        #             self.mask, self.image_res, imported_transforms
        #         )
        #     transforms.append(imported_transforms)
        original_size = image.GetSize()
        if spatial_preprocessing is not None:
            if (
                self.mask is not None
                and spatial_preprocessing.get("mask_to_bbox") is not None
            ):
                if spatial_preprocessing.get("mask_to_bbox") is True:
                    print("computing mask bounding box")
                    mask_bbox = compute_mask_to_bbox(self.mask)
                    spatial_preprocessing.update({"mask_bbox": mask_bbox})

            if spatial_preprocessing.get("mask_bbox") is not None:
                bbox = spatial_preprocessing["mask_bbox"]
                print("cropping to mask")
                translation_transform = gen_rigid_translation(
                    image, self.image_res, bbox[0], bbox[1], bbox[2], bbox[3]
                )

                (
                    composite_transform,
                    _,
                    final_tform,
                ) = prepare_wsireg_transform_data(
                    {"initial": [translation_transform]}
                )

                image = transform_plane(
                    image, final_tform, composite_transform
                )
                # image = transform_2d_image_itkelx(
                #     image, [translation_transform]
                # )

                self.original_size_transform = gen_rig_to_original(
                    original_size, deepcopy(translation_transform)
                )

                if self.mask is not None:
                    self.mask.SetSpacing((self.image_res, self.image_res))
                    # self.mask = transform_2d_image_itkelx(
                    #     self.mask, [translation_transform]
                    # )
                    self.mask = transform_plane(
                        self.mask, final_tform, composite_transform
                    )
                transforms.append(translation_transform)

            if spatial_preprocessing.get("rot_cc") is not None:
                rotangle = spatial_preprocessing["rot_cc"]
                if rotangle is not None and rotangle != 0:
                    print(f"rotating counter-clockwise {rotangle}")
                    rot_tform = gen_rigid_tform_rot(
                        image, self.image_res, rotangle
                    )
                    (
                        composite_transform,
                        _,
                        final_tform,
                    ) = prepare_wsireg_transform_data({"initial": [rot_tform]})

                    image = transform_plane(
                        image, final_tform, composite_transform
                    )
                    # image = transform_2d_image_itkelx(image, [rot_tform])

                    if self.mask is not None:
                        self.mask.SetSpacing((self.image_res, self.image_res))
                        # self.mask = transform_2d_image_itkelx(
                        #     self.mask, [translation_transform]
                        # )
                        self.mask = transform_plane(
                            self.mask, final_tform, composite_transform
                        )
                    transforms.append(rot_tform)

            if spatial_preprocessing.get("flip") is not None:
                flip_direction = spatial_preprocessing["flip"]
                if flip_direction != "None" and flip_direction is not None:
                    print(f"flipping image {flip_direction}")

                    flip_tform = gen_aff_tform_flip(
                        image, self.image_res, flip_direction
                    )

                    (
                        composite_transform,
                        _,
                        final_tform,
                    ) = prepare_wsireg_transform_data(
                        {"initial": [flip_tform]}
                    )

                    image = transform_plane(
                        image, final_tform, composite_transform
                    )

                    if self.mask is not None:
                        self.mask.SetSpacing((self.image_res, self.image_res))
                        self.mask = transform_plane(
                            self.mask, final_tform, composite_transform
                        )

                    transforms.append(flip_tform)

            downsampling = spatial_preprocessing.get("downsample")

            if downsampling is not None and downsampling > 1:
                print(
                    "performing downsampling by factor: {}".format(
                        downsampling
                    )
                )
                image.SetSpacing((self.image_res, self.image_res))
                image = sitk.Shrink(image, (downsampling, downsampling))

                if self.mask is not None:

                    self.mask.SetSpacing((self.image_res, self.image_res))
                    self.mask = sitk.Shrink(
                        self.mask, (downsampling, downsampling)
                    )
        if len(transforms) == 0:
            transforms = None
        if (
            spatial_preprocessing.get("use_mask") is not None
            and spatial_preprocessing.get("use_mask") is False
        ):
            self.mask = None
        return image, transforms

    def reg_image_sitk_to_itk(self, cast_to_float32=True):
        origin = self.reg_image.GetOrigin()
        spacing = self.reg_image.GetSpacing()
        # direction = image.GetDirection()
        is_vector = self.reg_image.GetNumberOfComponentsPerPixel() > 1
        if cast_to_float32 is True:
            self.reg_image = sitk.Cast(self.reg_image, sitk.sitkFloat32)
            self.reg_image = sitk.GetArrayFromImage(self.reg_image)
        else:
            self.reg_image = sitk.GetArrayFromImage(self.reg_image)

        self.reg_image = itk.GetImageFromArray(
            self.reg_image, is_vector=is_vector
        )
        self.reg_image.SetOrigin(origin)
        self.reg_image.SetSpacing(spacing)

        if self.mask is not None:
            origin = self.mask.GetOrigin()
            spacing = self.mask.GetSpacing()
            # direction = image.GetDirection()
            is_vector = self.mask.GetNumberOfComponentsPerPixel() > 1
            if cast_to_float32 is True:
                self.mask = sitk.Cast(self.mask, sitk.sitkFloat32)
                self.mask = sitk.GetArrayFromImage(self.mask)
            else:
                self.mask = sitk.GetArrayFromImage(self.mask)

            self.mask = itk.GetImageFromArray(self.mask, is_vector=is_vector)
            self.mask.SetOrigin(origin)
            self.mask.SetSpacing(spacing)

            mask_im_type = itk.Image[itk.UC, 2]
            self.mask = itk.binary_threshold_image_filter(
                self.mask,
                lower_threshold=1,
                inside_value=1,
                ttype=(type(self.mask), mask_im_type),
            )

    def transform_image(
        self,
        image_name,
        transform_data,
        file_writer="ome.tiff",
        output_dir="",
        **transformation_opts,
    ):
        if transform_data is not None:
            (
                itk_composite,
                itk_transforms,
                final_transform,
            ) = prepare_wsireg_transform_data(transform_data)
        else:
            itk_composite, itk_transforms, final_transform = None, None, None

        if file_writer.lower() == "ome.zarr" or file_writer.lower() == "zarr":
            im_fp = transform_to_ome_zarr(
                self, output_dir, **transformation_opts
            )

        if file_writer.lower() == "ome.tiff" or transform_data is None:
            ometiffwriter = OmeTiffWriter(self)
            im_fp = ometiffwriter.write_image_by_plane(
                image_name,
                output_dir=output_dir,
                final_transform=final_transform,
                composite_transform=itk_composite,
                **transformation_opts,
            )

        if file_writer.lower() == "ome.tiff-bytile":
            ometiffwriter = OmeTiffWriter(self)
            im_fp = ometiffwriter.write_image_by_tile(
                image_name,
                output_dir=output_dir,
                final_transform=final_transform,
                itk_transforms=itk_transforms,
                composite_transform=itk_composite,
                **transformation_opts,
            )
        return im_fp



from copy import deepcopy
import SimpleITK as sitk


def euler_elx_to_itk2d(tform):
    euler2d = sitk.Euler2DTransform()

    center = [float(p) for p in tform['CenterOfRotationPoint']]
    euler2d.SetFixedParameters(center)
    elx_parameters = [float(p) for p in tform['TransformParameters']]
    euler2d.SetParameters(elx_parameters)

    return euler2d


def similarity_elx_to_itk2d(tform):
    similarity2d = sitk.Similarity2DTransform()

    center = [float(p) for p in tform['CenterOfRotationPoint']]
    similarity2d.SetFixedParameters(center)
    elx_parameters = [float(p) for p in tform['TransformParameters']]
    similarity2d.SetParameters(elx_parameters)

    return similarity2d


def affine_elx_to_itk2d(tform):
    im_dimension = len(tform["Size"])
    affine2d = sitk.AffineTransform(im_dimension)

    center = [float(p) for p in tform['CenterOfRotationPoint']]
    affine2d.SetFixedParameters(center)
    elx_parameters = [float(p) for p in tform['TransformParameters']]
    affine2d.SetParameters(elx_parameters)

    return affine2d


def bspline_elx_to_itk2d(tform):
    im_dimension = len(tform["Size"])

    bspline2d = sitk.BSplineTransform(im_dimension, 3)
    bspline2d.SetTransformDomainOrigin(
        [float(p) for p in tform['Origin']]
    )  # from fixed image
    bspline2d.SetTransformDomainPhysicalDimensions(
        [int(p) for p in tform['Size']]
    )  # from fixed image
    bspline2d.SetTransformDomainDirection(
        [float(p) for p in tform['Direction']]
    )  # from fixed image

    fixedParams = [int(p) for p in tform['GridSize']]
    fixedParams += [float(p) for p in tform['GridOrigin']]
    fixedParams += [float(p) for p in tform['GridSpacing']]
    fixedParams += [float(p) for p in tform['GridDirection']]
    bspline2d.SetFixedParameters(fixedParams)
    bspline2d.SetParameters([float(p) for p in tform['TransformParameters']])
    return bspline2d


def convert_to_itk(tform):

    if tform["Transform"][0] == "AffineTransform":
        itk_tform = affine_elx_to_itk2d(tform)
    elif tform["Transform"][0] == "SimilarityTransform":
        itk_tform = similarity_elx_to_itk2d(tform)
    elif tform["Transform"][0] == "EulerTransform":
        itk_tform = euler_elx_to_itk2d(tform)
    elif tform["Transform"][0] == "BSplineTransform":
        itk_tform = bspline_elx_to_itk2d(tform)

    itk_tform.OutputSpacing = [float(p) for p in tform["Spacing"]]
    itk_tform.OutputDirection = [float(p) for p in tform["Direction"]]
    itk_tform.OutputOrigin = [float(p) for p in tform["Origin"]]
    itk_tform.OutputSize = [int(p) for p in tform["Size"]]
    itk_tform.ResampleInterpolator = tform["ResampleInterpolator"][0]

    return itk_tform


def get_elastix_transforms(transformations):
    elastix_transforms = deepcopy(transformations)

    for k, v in elastix_transforms.items():
        elastix_transforms.update({k: [t.elastix_transform for t in v]})

    return elastix_transforms



import numpy as np
from wsireg.reg_images.loader import reg_image_loader
from wsireg.reg_images.reg_image import RegImage
from wsireg.utils.tform_utils import (
    prepare_wsireg_transform_data,
)
from wsireg.utils.im_utils import (
    transform_to_ome_tiff_merge,
)


class MergeRegImage(RegImage):
    def __init__(
        self,
        image_fp,
        image_res,
        channel_names=None,
        channel_colors=None,
    ):

        if isinstance(image_fp, list) is False:
            raise ValueError(
                "MergeRegImage requires a list of images to merge"
            )

        if isinstance(image_res, list) is False:
            raise ValueError(
                "MergeRegImage requires a list of image resolutions for each image to merge"
            )

        if channel_names is None:
            channel_names = [None for i in range(0, len(image_fp))]

        if channel_colors is None:
            channel_colors = [None for i in range(0, len(image_fp))]

        images = []
        for im_idx, image_data in enumerate(
            zip(image_fp, image_res, channel_names, channel_colors)
        ):
            image, image_res, channel_names, channel_colors = image_data
            imdata = reg_image_loader(
                image,
                image_res,
                channel_names=channel_names,
                channel_colors=channel_colors,
            )
            if (
                imdata.channel_names is None
                or len(imdata.channel_names) != imdata.n_ch
            ):
                imdata.channel_names = [
                    f"C{idx}" for idx in range(0, imdata.n_ch)
                ]

            images.append(imdata)

        if all([im.im_dtype == images[0].im_dtype for im in images]) is False:
            raise ValueError(
                "MergeRegImage requires a list of images with matching data type"
            )

        if any([im.is_rgb for im in images]) is True:
            raise ValueError(
                "MergeRegImage does not support merger of RGB images"
            )

        self.images = images
        self.image_fps = image_fp
        self.im_dtype = self.images[0].im_dtype

        # rgb merging not supported!
        self.is_rgb = False

        self.n_ch = np.sum([i.n_ch for i in self.images])
        self.channel_names = [i.channel_names for i in self.images]
        self.original_size_transform = None

    def transform_image(
        self,
        image_name,
        sub_image_names,
        transform_data,
        file_writer="ome.tiff",
        output_dir="",
        **transformation_opts,
    ):
        if isinstance(sub_image_names, list) is False:
            raise ValueError(
                "MergeRegImage requires a list of image names for each image to merge"
            )

        if isinstance(transform_data, list) is False:
            raise ValueError(
                "MergeRegImage requires a list of image transforms for each image to merge"
            )

        if len(transform_data) != len(self.image_fps):
            raise ValueError(
                "MergeRegImage number of transforms does not match number of images"
            )

        def get_transform_data(transform_data):
            if transform_data is not None:
                (
                    itk_composite,
                    itk_transforms,
                    final_transform,
                ) = prepare_wsireg_transform_data(transform_data)
            else:
                itk_composite, itk_transforms, final_transform = (
                    None,
                    None,
                    None,
                )

            return itk_composite, itk_transforms, final_transform

        def prepare_channel_names(sub_image_name, channel_names):
            return [f"{sub_image_name} - {c}" for c in channel_names]

        self.channel_names = [
            prepare_channel_names(im_name, cnames)
            for im_name, cnames in zip(sub_image_names, self.channel_names)
        ]
        self.channel_names = [
            item for sublist in self.channel_names for item in sublist
        ]

        merge_transform_data = [get_transform_data(t) for t in transform_data]
        composite_transforms = [
            merge_transform_data[t][0]
            for t in range(len(merge_transform_data))
        ]
        final_transforms = [
            merge_transform_data[t][2]
            for t in range(len(merge_transform_data))
        ]

        spacing_check = [
            t.output_spacing for t in final_transforms if t is not None
        ]
        size_check = [t.output_size for t in final_transforms if t is not None]
        no_transform_check = [
            idx for idx, t in enumerate(final_transforms) if t is None
        ]

        for no_tform in no_transform_check:
            size_check.extend(
                [np.asarray(self.images[no_tform].im_dims)[1:].tolist()]
            )
            spacing_check.extend(
                [
                    [
                        self.images[no_tform].image_res,
                        self.images[no_tform].image_res,
                    ]
                ]
            )

        if all(spacing_check) is False:
            raise ValueError(
                "MergeRegImage all transforms output spacings and untransformed image spacings must match"
            )

        if all(size_check) is False:
            raise ValueError(
                "MergeRegImage all transforms output sizes and untransformed image sizes must match"
            )

        if file_writer.lower() == "ome.tiff":
            im_fp = transform_to_ome_tiff_merge(
                self,
                image_name,
                output_dir,
                final_transforms,
                composite_transforms,
                **transformation_opts,
            )
        return im_fp



from pathlib import Path
import json
import numpy as np
import itk
import SimpleITK as sitk
from wsireg.utils.reg_utils import json_to_pmap_dict
from wsireg.utils.itk_im_conversions import (
    sitk_image_to_itk_image,
    itk_image_to_sitk_image,
)
from wsireg.parameter_maps.transformations import (
    BASE_RIG_TFORM,
    BASE_AFF_TFORM,
)
from wsireg.reg_transform import RegTransform


NUMERIC_ELX_PARAMETERS = {
    "CenterOfRotationPoint": np.float64,
    "DefaultPixelValue": np.float64,
    "Direction": np.float64,
    "FixedImageDimension": np.int64,
    "Index": np.int64,
    "MovingImageDimension": np.int64,
    "NumberOfParameters": np.int64,
    "Origin": np.float64,
    "Size": np.int64,
    "Spacing": np.float64,
    "TransformParameters": np.float64,
}

ELX_LINEAR_TRANSFORMS = [
    "AffineTransform",
    "EulerTransform",
    "SimilarityTransform",
]

ELX_TO_ITK_INTERPOLATORS = {
    "FinalNearestNeighborInterpolator": sitk.sitkNearestNeighbor,
    "FinalLinearInterpolator": sitk.sitkLinear,
    "FinalBSplineInterpolator": sitk.sitkBSpline,
}


def prepare_tform_dict(tform_dict, shape_tform=False):

    tform_dict_out = {}
    for k, v in tform_dict.items():
        if k == "initial":
            tform_dict_out["initial"] = v
        else:
            tforms = []
            for tform in v:
                if "invert" in list(tform.keys()):
                    if shape_tform is False:
                        tforms.append(tform["image"])
                    else:
                        tforms.append(tform["invert"])
                else:
                    tforms.append(tform)
            tform_dict_out[k] = tforms

    return tform_dict_out


def transform_2d_image_itkelx(
    image, transformation_maps, writer="sitk", **zarr_kwargs
):
    """
    Transform 2D images with multiple models and return the transformed image
    or write the transformed image to disk as a .tif file.
    Multichannel or multicomponent images (RGB) have to be transformed a single channel at a time
    This function takes care of performing those transformations and reconstructing the image in the same
    data type as the input
    Parameters
    ----------
    image : SimpleITK.Image
        Image to be transformed
    transformation_maps : list
        list of SimpleElastix ParameterMaps to used for transformation
    Returns
    -------
    Transformed SimpleITK.Image
    """
    if transformation_maps is not None:
        tfx = itk.TransformixFilter.New()

        # TODO: add mask cropping here later

        #     print("mask cropping")
        #     tmap = sitk.ReadParameterFile(transformation_maps[0])
        #     x_min = int(float(tmap["MinimumX"][0]))
        #     x_max = int(float(tmap["MaximumX"][0]))
        #     y_min = int(float(tmap["MinimumY"][0]))
        #     y_max = int(float(tmap["MaximumY"][0]))
        #     image = image[x_min:x_max, y_min:y_max]
        #     origin = np.repeat(0, len(image.GetSize()))
        #     image.SetOrigin(tuple([int(i) for i in origin]))

        # else:
        transform_pobj = itk.ParameterObject.New()
        for idx, tmap in enumerate(transformation_maps):
            if isinstance(tmap, str):
                tmap = sitk.ReadParameterFile(tmap)

            if idx == 0:
                tmap["InitialTransformParametersFileName"] = (
                    "NoInitialTransform",
                )
                transform_pobj.AddParameterMap(tmap)
            else:
                tmap["InitialTransformParametersFileName"] = (
                    "NoInitialTransform",
                )
                transform_pobj.AddParameterMap(tmap)
        tfx.SetTransformParameterObject(transform_pobj)
        tfx.LogToConsoleOn()
        tfx.LogToFileOff()
    else:
        tfx = None

    # if tfx is None:
    #     xy_final_size = np.array(image.GetSize(), dtype=np.uint32)
    # else:
    #     xy_final_size = np.array(
    #         transformation_maps[-1]["Size"], dtype=np.uint32
    #     )

    if writer == "sitk" or writer is None:
        return transform_image_itkelx_to_sitk(image, tfx)
    elif writer == "zarr":
        return
    else:
        raise ValueError("writer type {} not recognized".format(writer))


def transform_image_to_sitk(image, tfx):

    # manage transformation/casting if data is multichannel or RGB
    # data is always returned in the same PixelIDType as it is entered

    pixel_id = image.GetPixelID()
    if tfx is not None:
        if pixel_id in list(range(1, 13)) and image.GetDepth() == 0:
            tfx.SetMovingImage(image)
            image = tfx.Execute()
            image = sitk.Cast(image, pixel_id)

        elif pixel_id in list(range(1, 13)) and image.GetDepth() > 0:
            images = []
            for chan in range(image.GetDepth()):
                tfx.SetMovingImage(image[:, :, chan])
                images.append(sitk.Cast(tfx.Execute(), pixel_id))
            image = sitk.JoinSeries(images)
            image = sitk.Cast(image, pixel_id)

        elif pixel_id > 12:
            images = []
            for idx in range(image.GetNumberOfComponentsPerPixel()):
                im = sitk.VectorIndexSelectionCast(image, idx)
                pixel_id_nonvec = im.GetPixelID()
                tfx.SetMovingImage(im)
                images.append(sitk.Cast(tfx.Execute(), pixel_id_nonvec))
                del im

            image = sitk.Compose(images)
            image = sitk.Cast(image, pixel_id)

    return image


def transform_image_itkelx_to_sitk(image, tfx):

    # manage transformation/casting if data is multichannel or RGB
    # data is always returned in the same PixelIDType as it is entered

    pixel_id = image.GetPixelID()
    if tfx is not None:
        if pixel_id in list(range(1, 13)) and image.GetDepth() == 0:
            image = sitk_image_to_itk_image(image, cast_to_float32=True)
            tfx.SetMovingImage(image)
            tfx.UpdateLargestPossibleRegion()
            image = tfx.GetOutput()
            image = itk_image_to_sitk_image(image)
            image = sitk.Cast(image, pixel_id)

        elif pixel_id in list(range(1, 13)) and image.GetDepth() > 0:
            images = []
            for chan in range(image.GetDepth()):
                image = sitk_image_to_itk_image(
                    image[:, :, chan], cast_to_float32=True
                )
                tfx.SetMovingImage(image)
                tfx.UpdateLargestPossibleRegion()
                image = tfx.GetOutput()
                image = itk_image_to_sitk_image(image)
                image = sitk.Cast(image, pixel_id)
                images.append(image)
            image = sitk.JoinSeries(images)
            image = sitk.Cast(image, pixel_id)

        elif pixel_id > 12:
            images = []
            for idx in range(image.GetNumberOfComponentsPerPixel()):
                im = sitk.VectorIndexSelectionCast(image, idx)
                pixel_id_nonvec = im.GetPixelID()
                im = sitk_image_to_itk_image(im, cast_to_float32=True)
                tfx.SetMovingImage(im)
                tfx.UpdateLargestPossibleRegion()
                im = tfx.GetOutput()
                im = itk_image_to_sitk_image(im)
                im = sitk.Cast(im, pixel_id_nonvec)
                images.append(im)
                del im

            image = sitk.Compose(images)
            image = sitk.Cast(image, pixel_id)

    return image


def apply_transform_dict_itkelx(
    image_fp,
    image_res,
    tform_dict_in,
    prepro_dict=None,
    is_shape_mask=False,
    writer="sitk",
    **im_tform_kwargs,
):
    """
    Apply a complex series of transformations in a python dictionary to an image

    Parameters
    ----------
    image_fp : str
        file path to the image to be transformed, it will be read in it's entirety
    image_res : float
        pixel resolution of image to be transformed
    tform_dict : dict of lists
        dict of SimpleElastix transformations stored in lists, may contain an "initial" transforms (preprocessing transforms)
        these will be applied first, then the key order of the dict will determine the rest of the transformations
    is_shape_mask : bool
        whether the image being transformed is a shape mask (determines import)

    Returns
    -------
    image: itk.Image
        image that has been transformed

    """

    if is_shape_mask is False:
        if isinstance(image_fp, sitk.Image):
            image = image_fp
        # else:
        #     image = RegImage(
        #         image_fp, image_res, prepro_dict=prepro_dict
        #     ).image
    else:
        image = sitk.GetImageFromArray(image_fp)
        del image_fp
        image.SetSpacing((image_res, image_res))

    if tform_dict_in is None:
        if writer == "zarr":
            image = transform_2d_image_itkelx(
                image,
                None,
                writer="zarr",
                zarr_store_dir=im_tform_kwargs["zarr_store_dir"],
                channel_names=im_tform_kwargs["channel_names"],
                channel_colors=im_tform_kwargs["channel_colors"],
            )
        else:
            image = transform_2d_image_itkelx(image, None)

    else:
        tform_dict = tform_dict_in.copy()

        if tform_dict.get("registered") is None and tform_dict.get(0) is None:
            tform_dict["registered"] = tform_dict["initial"]
            tform_dict.pop("initial", None)

            if isinstance(tform_dict.get("registered"), list) is False:
                tform_dict["registered"] = [tform_dict["registered"]]

            for idx in range(len(tform_dict["registered"])):
                tform_dict[idx] = [tform_dict["registered"][idx]]

            tform_dict.pop("registered", None)
        else:
            tform_dict = prepare_tform_dict(tform_dict, shape_tform=False)

        if "initial" in tform_dict:
            for initial_tform in tform_dict["initial"]:
                if isinstance(initial_tform, list) is False:
                    initial_tform = [initial_tform]

                for tform in initial_tform:
                    image = transform_2d_image_itkelx(image, [tform])

            tform_dict.pop("initial", None)

        for k, v in tform_dict.items():
            if writer == "zarr" and k == list(tform_dict.keys())[-1]:
                image = transform_2d_image_itkelx(
                    image,
                    v,
                    writer="zarr",
                    zarr_store_dir=im_tform_kwargs["zarr_store_dir"],
                    channel_names=im_tform_kwargs["channel_names"],
                    channel_colors=im_tform_kwargs["channel_colors"],
                )
            else:
                image = transform_2d_image_itkelx(image, v)

    return image


def compute_rot_bound(image, angle=30):
    """
    compute the bounds of an image after by an angle

    Parameters
    ----------
    image : sitk.Image
        SimpleITK image that will be rotated    angle
    angle : float
        angle of rotation in degrees, rotates counter-clockwise if positive

    Returns
    -------
    tuple of the rotated image's size in x and y

    """
    w, h = image.GetSize()[0], image.GetSize()[1]

    theta = np.radians(angle)
    c, s = np.abs(np.cos(theta)), np.abs(np.sin(theta))
    bound_w = (h * s) + (w * c)
    bound_h = (h * c) + (w * s)

    return bound_w, bound_h


def gen_rigid_tform_rot(image, spacing, angle):
    """
    generate a SimpleElastix transformation parameter Map to rotate image by angle
    Parameters
    ----------
    image : sitk.Image
        SimpleITK image that will be rotated
    spacing : float
        Physical spacing of the SimpleITK image
    angle : float
        angle of rotation in degrees, rotates counter-clockwise if positive

    Returns
    -------
    SimpleITK.ParameterMap of rotation transformation (EulerTransform)
    """
    tform = BASE_RIG_TFORM.copy()
    image.SetSpacing((spacing, spacing))
    bound_w, bound_h = compute_rot_bound(image, angle=angle)

    rot_cent_pt = image.TransformContinuousIndexToPhysicalPoint(
        ((bound_w - 1) / 2, (bound_h - 1) / 2)
    )

    c_x, c_y = (image.GetSize()[0] - 1) / 2, (image.GetSize()[1] - 1) / 2
    c_x_phy, c_y_phy = image.TransformContinuousIndexToPhysicalPoint(
        (c_x, c_y)
    )
    t_x = rot_cent_pt[0] - c_x_phy
    t_y = rot_cent_pt[1] - c_y_phy

    tform["Spacing"] = [str(spacing), str(spacing)]
    tform["Size"] = [str(int(np.ceil(bound_w))), str(int(np.ceil(bound_h)))]
    tform["CenterOfRotationPoint"] = [str(rot_cent_pt[0]), str(rot_cent_pt[1])]
    tform["TransformParameters"] = [
        str(np.radians(angle)),
        str(-1 * t_x),
        str(-1 * t_y),
    ]

    return tform


def gen_rigid_translation(
    image, spacing, translation_x, translation_y, size_x, size_y
):
    """
    generate a SimpleElastix transformation parameter Map to rotate image by angle
    Parameters
    ----------
    image : sitk.Image
        SimpleITK image that will be rotated
    spacing : float
        Physical spacing of the SimpleITK image

    Returns
    -------
    SimpleITK.ParameterMap of rotation transformation (EulerTransform)
    """
    tform = BASE_RIG_TFORM.copy()
    image.SetSpacing((spacing, spacing))
    bound_w, bound_h = compute_rot_bound(image, angle=0)

    rot_cent_pt = image.TransformContinuousIndexToPhysicalPoint(
        ((bound_w - 1) / 2, (bound_h - 1) / 2)
    )
    (
        translation_x,
        translation_y,
    ) = image.TransformContinuousIndexToPhysicalPoint(
        (float(translation_x), float(translation_y))
    )
    # c_x, c_y = (image.GetSize()[0] - 1) / 2, (image.GetSize()[1] - 1) / 2

    tform["Spacing"] = [str(spacing), str(spacing)]
    tform["Size"] = [str(size_x), str(size_y)]
    tform["CenterOfRotationPoint"] = [str(rot_cent_pt[0]), str(rot_cent_pt[1])]
    tform["TransformParameters"] = [
        str(0),
        str(translation_x),
        str(translation_y),
    ]

    return tform


def gen_rig_to_original(original_size, crop_transform):
    crop_transform["Size"] = [str(original_size[0]), str(original_size[1])]
    tform_params = [float(t) for t in crop_transform["TransformParameters"]]
    crop_transform["TransformParameters"] = [
        str(0),
        str(tform_params[1] * -1),
        str(tform_params[2] * -1),
    ]
    return crop_transform


def gen_aff_tform_flip(image, spacing, flip="h"):
    """
    generate a SimpleElastix transformation parameter Map to horizontally or vertically flip image

    Parameters
    ----------
    image : sitk.Image
        SimpleITK image that will be rotated
    spacing : float
        Physical spacing of the SimpleITK image
    flip : str
        "h" or "v" for horizontal or vertical flipping, respectively

    Returns
    -------
    SimpleITK.ParameterMap of flipping transformation (AffineTransform)

    """
    tform = BASE_AFF_TFORM.copy()
    image.SetSpacing((spacing, spacing))
    bound_w, bound_h = compute_rot_bound(image, angle=0)
    rot_cent_pt = image.TransformContinuousIndexToPhysicalPoint(
        ((bound_w - 1) / 2, (bound_h - 1) / 2)
    )

    tform["Spacing"] = [str(spacing), str(spacing)]
    tform["Size"] = [str(int(bound_w)), str(int(bound_h))]

    tform["CenterOfRotationPoint"] = [str(rot_cent_pt[0]), str(rot_cent_pt[1])]
    if flip == "h":
        tform_params = ["-1", "0", "0", "1", "0", "0"]
    elif flip == "v":
        tform_params = ["1", "0", "0", "-1", "0", "0"]

    tform["TransformParameters"] = tform_params

    return tform


def make_composite_itk(itk_tforms):

    itk_composite = sitk.CompositeTransform(2)
    for t in itk_tforms:
        itk_composite.AddTransform(t.itk_transform)
    return itk_composite


def get_final_tform(parameter_data):
    if (
        isinstance(parameter_data, str)
        and Path(parameter_data).suffix == ".json"
    ):
        parameter_data = json.load(open(parameter_data, "r"))

    final_key = list(parameter_data.keys())[-1]

    final_tform = parameter_data[final_key][-1]
    return final_tform


def collate_wsireg_transforms(parameter_data):

    if (
        isinstance(parameter_data, str)
        and Path(parameter_data).suffix == ".json"
    ):
        parameter_data = json.load(open(parameter_data, "r"))

    parameter_data_list = []
    for k, v in parameter_data.items():
        if k == "initial":
            if isinstance(v, dict):
                parameter_data_list.append([v])
            elif isinstance(v, list):
                for init_tform in v:
                    parameter_data_list.append([init_tform])
        else:
            sub_tform = []
            if isinstance(v, dict):
                sub_tform.append(v)
            elif isinstance(v, list):
                sub_tform += v
            sub_tform = sub_tform[::-1]
            parameter_data_list.append(sub_tform)

    flat_pmap_list = [
        item for sublist in parameter_data_list for item in sublist
    ]

    if all([isinstance(t, dict) for t in flat_pmap_list]):
        flat_pmap_list = [RegTransform(t) for t in flat_pmap_list]

    return flat_pmap_list


def wsireg_transforms_to_itk_composite(parameter_data):

    reg_transforms = collate_wsireg_transforms(parameter_data)
    composite_tform = make_composite_itk(reg_transforms)

    return composite_tform, reg_transforms


def prepare_wsireg_transform_data(transform_data):
    if isinstance(transform_data, str) is True:
        transform_data = json_to_pmap_dict(transform_data)

    if transform_data is not None:
        (
            composite_transform,
            itk_transforms,
        ) = wsireg_transforms_to_itk_composite(transform_data)
        final_tform = itk_transforms[-1]
    return composite_transform, itk_transforms, final_tform


def wsireg_transforms_to_resampler(final_tform):
    resampler = sitk.ResampleImageFilter()
    resampler.SetOutputOrigin(final_tform.output_origin)
    resampler.SetSize(final_tform.output_size)
    resampler.SetOutputDirection(final_tform.output_direction)
    resampler.SetOutputSpacing(final_tform.output_spacing)
    interpolator = ELX_TO_ITK_INTERPOLATORS.get(
        final_tform.resample_interpolator
    )
    resampler.SetInterpolator(interpolator)
    return resampler


def sitk_transform_image(image, final_tform, composite_transform):
    resampler = wsireg_transforms_to_resampler(final_tform)
    resampler.SetTransform(composite_transform)
    image = resampler.Execute(image)
    return image


#
# def apply_transform_dict(
#     image_fp,
#     image_res,
#     tform_dict_in,
#     prepro_dict=None,
#     is_shape_mask=False,
#     writer="sitk",
#     **im_tform_kwargs,
# ):
#     """
#     Apply a complex series of transformations in a python dictionary to an image
#
#     Parameters
#     ----------
#     image_fp : str
#         file path to the image to be transformed, it will be read in it's entirety
#     image_res : float
#         pixel resolution of image to be transformed
#     tform_dict : dict of lists
#         dict of SimpleElastix transformations stored in lists, may contain an "initial" transforms (preprocessing transforms)
#         these will be applied first, then the key order of the dict will determine the rest of the transformations
#     prepro_dict : dict
#         preprocessing to perform on image before transformation, default None reads full image
#     is_shape_mask : bool
#         whether the image being transformed is a shape mask (determines import)
#
#     Returns
#     -------
#     image: itk.Image
#         image that has been transformed
#
#     """
#
#     if is_shape_mask is False:
#         if isinstance(image_fp, sitk.Image):
#             image = image_fp
#         # else:
#         #     image = RegImage(
#         #         image_fp, image_res, prepro_dict=prepro_dict
#         #     ).image
#     else:
#         image = sitk.GetImageFromArray(image_fp)
#         del image_fp
#         image.SetSpacing((image_res, image_res))
#
#     if tform_dict_in is None:
#         if writer == "zarr":
#             image = transform_2d_image(
#                 image,
#                 None,
#                 writer="zarr",
#                 zarr_store_dir=im_tform_kwargs["zarr_store_dir"],
#                 channel_names=im_tform_kwargs["channel_names"],
#                 channel_colors=im_tform_kwargs["channel_colors"],
#             )
#         else:
#             image = transform_2d_image(image, None)
#
#     else:
#         tform_dict = tform_dict_in.copy()
#
#         if tform_dict.get("registered") is None and tform_dict.get(0) is None:
#             tform_dict["registered"] = tform_dict["initial"]
#             tform_dict.pop("initial", None)
#
#             if isinstance(tform_dict.get("registered"), list) is False:
#                 tform_dict["registered"] = [tform_dict["registered"]]
#
#             for idx in range(len(tform_dict["registered"])):
#                 tform_dict[idx] = [tform_dict["registered"][idx]]
#
#             tform_dict.pop("registered", None)
#         else:
#             tform_dict = prepare_tform_dict(tform_dict, shape_tform=False)
#
#         if "initial" in tform_dict:
#             for initial_tform in tform_dict["initial"]:
#                 if isinstance(initial_tform, list) is False:
#                     initial_tform = [initial_tform]
#
#                 for tform in initial_tform:
#                     image = transform_2d_image(image, [tform])
#
#             tform_dict.pop("initial", None)
#
#         for k, v in tform_dict.items():
#             if writer == "zarr" and k == list(tform_dict.keys())[-1]:
#                 image = transform_2d_image(
#                     image,
#                     v,
#                     writer="zarr",
#                     zarr_store_dir=im_tform_kwargs["zarr_store_dir"],
#                     channel_names=im_tform_kwargs["channel_names"],
#                     channel_colors=im_tform_kwargs["channel_colors"],
#                 )
#             else:
#                 image = transform_2d_image(image, v)
#
#     return image


#
# def transform_2d_image(
#     image, transformation_maps, writer="sitk", **zarr_kwargs
# ):
#     """
#     Transform 2D images with multiple models and return the transformed image
#     or write the transformed image to disk as a .tif file.
#     Multichannel or multicomponent images (RGB) have to be transformed a single channel at a time
#     This function takes care of performing those transformations and reconstructing the image in the same
#     data type as the input
#     Parameters
#     ----------
#     image : SimpleITK.Image
#         Image to be transformed
#     transformation_maps : list
#         list of SimpleElastix ParameterMaps to used for transformation
#     Returns
#     -------
#     Transformed SimpleITK.Image
#     """
#     if transformation_maps is not None:
#
#         try:
#             tfx = sitk.TransformixImageFilter()
#         except AttributeError:
#             tfx = sitk.SimpleTransformix()
#
#         # TODO: add mask cropping here later
#
#         #     print("mask cropping")
#         #     tmap = sitk.ReadParameterFile(transformation_maps[0])
#         #     x_min = int(float(tmap["MinimumX"][0]))
#         #     x_max = int(float(tmap["MaximumX"][0]))
#         #     y_min = int(float(tmap["MinimumY"][0]))
#         #     y_max = int(float(tmap["MaximumY"][0]))
#         #     image = image[x_min:x_max, y_min:y_max]
#         #     origin = np.repeat(0, len(image.GetSize()))
#         #     image.SetOrigin(tuple([int(i) for i in origin]))
#
#         # else:
#         for idx, tmap in enumerate(transformation_maps):
#             if isinstance(tmap, str):
#                 tmap = sitk.ReadParameterFile(tmap)
#
#             if idx == 0:
#                 tmap["InitialTransformParametersFileName"] = (
#                     "NoInitialTransform",
#                 )
#                 tfx.SetTransformParameterMap(tmap)
#             else:
#                 tmap["InitialTransformParametersFileName"] = (
#                     "NoInitialTransform",
#                 )
#
#                 tfx.AddTransformParameterMap(tmap)
#         tfx.LogToConsoleOn()
#         tfx.LogToFileOff()
#     else:
#         tfx = None
#
#     # if tfx is None:
#     #     xy_final_size = np.array(image.GetSize(), dtype=np.uint32)
#     # else:
#     #     xy_final_size = np.array(
#     #         transformation_maps[-1]["Size"], dtype=np.uint32
#     #     )
#
#     if writer == "sitk" or writer is None:
#         return transform_image_to_sitk(image, tfx)
#     elif writer == "zarr":
#         return
#     else:
#         raise ValueError("writer type {} not recognized".format(writer))
#



DEFAULT_REG_PARAM_MAPS = {
    'rigid': {
        "AutomaticScalesEstimation": ['true'],
        "AutomaticTransformInitialization": ['true'],
        "BSplineInterpolationOrder": ['1'],
        "CompressResultImage": ['true'],
        "DefaultPixelValue": ['0'],
        "ErodeMask": ['false'],
        "FinalBSplineInterpolationOrder": ['1'],
        "FixedImageDimension": ['2'],
        "FixedImagePyramid": ['FixedRecursiveImagePyramid'],
        "FixedInternalImagePixelType": ['float'],
        "HowToCombineTransforms": ['Compose'],
        "ImageSampler": ["Random"],
        "Interpolator": ['LinearInterpolator'],
        "MaximumNumberOfIterations": ['200'],
        "MaximumNumberOfSamplingAttempts": [
            '10',
        ],
        "MaximumStepLength": [
            '100.0',
            '75.0',
            '66.0',
            '50.0',
            '25.0',
            '15.0',
            '10.0',
            '10.0',
            '5.0',
            '1.0',
        ],
        "Metric": ['AdvancedMattesMutualInformation'],
        "MovingImageDimension": ['2'],
        "MovingImagePyramid": ['MovingRecursiveImagePyramid'],
        "MovingInternalImagePixelType": ['float'],
        "NewSamplesEveryIteration": ['true'],
        "NumberOfHistogramBins": ['16'],
        "NumberOfResolutions": ['10'],
        "NumberOfSpatialSamples": ['10000'],
        "Optimizer": ['AdaptiveStochasticGradientDescent'],
        "Registration": ['MultiResolutionRegistration'],
        "RequiredRatioOfValidSamples": ['0.05'],
        "ResampleInterpolator": ['FinalNearestNeighborInterpolator'],
        "Resampler": ['DefaultResampler'],
        "ResultImageFormat": ['mha'],
        "ResultImagePixelType": ['short'],
        "Transform": ['EulerTransform'],
        "UseDirectionCosines": ['true'],
        "WriteResultImage": ['false'],
        "WriteTransformParametersEachResolution": ['true'],
    },
    'affine': {
        "AutomaticScalesEstimation": ['true'],
        "AutomaticTransformInitialization": ['true'],
        "BSplineInterpolationOrder": ['1'],
        "CompressResultImage": ['true'],
        "DefaultPixelValue": ['0'],
        "ErodeMask": ['false'],
        "FinalBSplineInterpolationOrder": ['1'],
        "FixedImageDimension": ['2'],
        "FixedImagePyramid": ['FixedRecursiveImagePyramid'],
        "FixedInternalImagePixelType": ['float'],
        "HowToCombineTransforms": ['Compose'],
        "ImageSampler": ['Random'],
        "Interpolator": ['LinearInterpolator'],
        "MaximumNumberOfIterations": ['200'],
        "MaximumNumberOfSamplingAttempts": [
            '10',
        ],
        "MaximumStepLength": [
            '100.0',
            '75.0',
            '66.0',
            '50.0',
            '25.0',
            '15.0',
            '10.0',
            '10.0',
            '5.0',
            '1.0',
        ],
        "Metric": ['AdvancedMattesMutualInformation'],
        "MovingImageDimension": ['2'],
        "MovingImagePyramid": ['MovingRecursiveImagePyramid'],
        "MovingInternalImagePixelType": ['float'],
        "NewSamplesEveryIteration": ['true'],
        "NumberOfHistogramBins": ['32'],
        "NumberOfResolutions": ['10'],
        "NumberOfSpatialSamples": ['10000'],
        "Optimizer": ['AdaptiveStochasticGradientDescent'],
        "Registration": ['MultiResolutionRegistration'],
        "RequiredRatioOfValidSamples": ['0.05'],
        "ResampleInterpolator": ['FinalNearestNeighborInterpolator'],
        "Resampler": ['DefaultResampler'],
        "ResultImageFormat": ['mha'],
        "ResultImagePixelType": ['short'],
        "Transform": ['AffineTransform'],
        "UseDirectionCosines": ['true'],
        "WriteResultImage": ['false'],
        "WriteTransformParametersEachResolution": ['true'],
    },
    'similarity': {
        "AutomaticScalesEstimation": ['true'],
        "AutomaticTransformInitialization": ['true'],
        "BSplineInterpolationOrder": ['1'],
        "CompressResultImage": ['true'],
        "DefaultPixelValue": ['0'],
        "ErodeMask": ['false'],
        "FinalBSplineInterpolationOrder": ['1'],
        "FixedImageDimension": ['2'],
        "FixedImagePyramid": ['FixedRecursiveImagePyramid'],
        "FixedInternalImagePixelType": ['float'],
        "HowToCombineTransforms": ['Compose'],
        "ImageSampler": ['Random'],
        "Interpolator": ['LinearInterpolator'],
        "MaximumNumberOfIterations": ['200'],
        "MaximumNumberOfSamplingAttempts": [
            '10',
        ],
        "MaximumStepLength": [
            '100.0',
            '75.0',
            '66.0',
            '50.0',
            '25.0',
            '15.0',
            '10.0',
            '10.0',
            '5.0',
            '1.0',
        ],
        "Metric": ['AdvancedMattesMutualInformation'],
        "MovingImageDimension": ['2'],
        "MovingImagePyramid": ['MovingRecursiveImagePyramid'],
        "MovingInternalImagePixelType": ['float'],
        "NewSamplesEveryIteration": ['true'],
        "NumberOfHistogramBins": ['32'],
        "NumberOfResolutions": ['10'],
        "NumberOfSpatialSamples": ['10000'],
        "Optimizer": ['AdaptiveStochasticGradientDescent'],
        "Registration": ['MultiResolutionRegistration'],
        "RequiredRatioOfValidSamples": ['0.05'],
        "ResampleInterpolator": ['FinalNearestNeighborInterpolator'],
        "Resampler": ['DefaultResampler'],
        "ResultImageFormat": ['mha'],
        "ResultImagePixelType": ['short'],
        "Transform": ['SimilarityTransform'],
        "UseDirectionCosines": ['true'],
        "WriteResultImage": ['false'],
        "WriteTransformParametersEachResolution": ['true'],
    },
    'nl': {
        "AutomaticScalesEstimation": ['true'],
        "AutomaticTransformInitialization": ['false'],
        "BSplineInterpolationOrder": ['1'],
        "CompressResultImage": ['true'],
        "DefaultPixelValue": ['0'],
        "ErodeMask": ['false'],
        "FinalBSplineInterpolationOrder": ['1'],
        "FinalGridSpacingInPhysicalUnits": ['100'],
        "FixedImageDimension": ['2'],
        "FixedImagePyramid": ['FixedRecursiveImagePyramid'],
        "FixedInternalImagePixelType": ['float'],
        "GridSpacingSchedule": [
            '512',
            '512',
            '392',
            '392',
            '256',
            '256',
            '128',
            '128',
            '64',
            '64',
            '32',
            '32',
            '16',
            '16',
            '4',
            '4',
            '2',
            '2',
            '1',
            '1',
        ],
        "HowToCombineTransforms": ['Compose'],
        "ImageSampler": ['Random'],
        "Interpolator": ['LinearInterpolator'],
        "MaximumNumberOfIterations": ['200'],
        "MaximumNumberOfSamplingAttempts": [
            '10',
        ],
        "MaximumStepLength": [
            '100',
            '90',
            '70',
            '50',
            '40',
            '30',
            '20',
            '10',
            '1',
            '1',
        ],
        "Metric": ['AdvancedMattesMutualInformation'],
        "MovingImageDimension": ['2'],
        "MovingImagePyramid": ['MovingRecursiveImagePyramid'],
        "MovingInternalImagePixelType": ['float'],
        "NewSamplesEveryIteration": ['true'],
        "NumberOfHistogramBins": ['32'],
        "NumberOfResolutions": ['10'],
        "NumberOfSpatialSamples": ['50000'],
        "Optimizer": ['AdaptiveStochasticGradientDescent'],
        "Registration": ['MultiResolutionRegistration'],
        "RequiredRatioOfValidSamples": ['0.05'],
        "ResampleInterpolator": ['FinalNearestNeighborInterpolator'],
        "Resampler": ['DefaultResampler'],
        "ResultImageFormat": ['mha'],
        "ResultImagePixelType": ['short'],
        "Transform": ['BSplineTransform'],
        "UseDirectionCosines": ['true'],
        "WriteResultImage": ['false'],
        "WriteTransformParametersEachResolution": ['true'],
    },
    'fi_correction': {
        "AutomaticScalesEstimation": ['true'],
        "AutomaticTransformInitialization": ['false'],
        "BSplineInterpolationOrder": ['1'],
        "CompressResultImage": ['true'],
        "DefaultPixelValue": ['0'],
        "ErodeMask": ['false'],
        "FinalBSplineInterpolationOrder": ['1'],
        "FixedImageDimension": ['2'],
        "FixedImagePyramid": ['FixedRecursiveImagePyramid'],
        "FixedInternalImagePixelType": ['float'],
        "HowToCombineTransforms": ['Compose'],
        "ImagePyramidSchedule": ['8', '8', '4', '4', '2', '2', '1', '1'],
        "ImageSampler": ['Random'],
        "Interpolator": ['LinearInterpolator'],
        "MaximumNumberOfIterations": ['75'],
        "MaximumNumberOfSamplingAttempts": [
            '10',
        ],
        "MaximumStepLength": ['100', '50', '20', '10'],
        "Metric": ['AdvancedMattesMutualInformation'],
        "MovingImageDimension": ['2'],
        "MovingImagePyramid": ['MovingRecursiveImagePyramid'],
        "MovingInternalImagePixelType": ['float'],
        "NewSamplesEveryIteration": ['true'],
        "NumberOfHistogramBins": ['16'],
        "NumberOfResolutions": ['4'],
        "NumberOfSpatialSamples": ['10000'],
        "Optimizer": ['AdaptiveStochasticGradientDescent'],
        "Registration": ['MultiResolutionRegistration'],
        "RequiredRatioOfValidSamples": ['0.05'],
        "ResampleInterpolator": ['FinalNearestNeighborInterpolator'],
        "Resampler": ['DefaultResampler'],
        "ResultImageFormat": ['mha'],
        "ResultImagePixelType": ['short'],
        "Transform": ['EulerTransform'],
        "UseDirectionCosines": ['true'],
        "WriteResultImage": ['true'],
        "WriteTransformParametersEachResolution": ['false'],
    },
}

test_rig = DEFAULT_REG_PARAM_MAPS["rigid"].copy()
test_aff = DEFAULT_REG_PARAM_MAPS["affine"].copy()
test_sim = DEFAULT_REG_PARAM_MAPS["similarity"].copy()
test_nl = DEFAULT_REG_PARAM_MAPS["nl"].copy()

test_rig["MaximumNumberOfIterations"] = ["10"]
test_aff["MaximumNumberOfIterations"] = ["10"]
test_sim["MaximumNumberOfIterations"] = ["10"]
test_nl["MaximumNumberOfIterations"] = ["10"]

DEFAULT_REG_PARAM_MAPS["rigid_test"] = test_rig
DEFAULT_REG_PARAM_MAPS["affine_test"] = test_aff
DEFAULT_REG_PARAM_MAPS["similarity_test"] = test_sim
DEFAULT_REG_PARAM_MAPS["nl_test"] = test_nl

# advanced mean squares
ams_rig = DEFAULT_REG_PARAM_MAPS["rigid"].copy()
ams_aff = DEFAULT_REG_PARAM_MAPS["affine"].copy()
ams_sim = DEFAULT_REG_PARAM_MAPS["similarity"].copy()
ams_nl = DEFAULT_REG_PARAM_MAPS["nl"].copy()

ams_rig["Metric"] = ["AdvancedMeanSquares"]
ams_aff["Metric"] = ["AdvancedMeanSquares"]
ams_sim["Metric"] = ["AdvancedMeanSquares"]
ams_nl["Metric"] = ["AdvancedMeanSquares"]

DEFAULT_REG_PARAM_MAPS["rigid_ams"] = ams_rig
DEFAULT_REG_PARAM_MAPS["affine_ams"] = ams_aff
DEFAULT_REG_PARAM_MAPS["similarity_ams"] = ams_sim
DEFAULT_REG_PARAM_MAPS["nl_ams"] = ams_nl

# normalized correlation
anc_rig = DEFAULT_REG_PARAM_MAPS["rigid"].copy()
anc_aff = DEFAULT_REG_PARAM_MAPS["affine"].copy()
anc_sim = DEFAULT_REG_PARAM_MAPS["similarity"].copy()
anc_nl = DEFAULT_REG_PARAM_MAPS["nl"].copy()

anc_rig["Metric"] = ["AdvancedNormalizedCorrelation"]
anc_aff["Metric"] = ["AdvancedNormalizedCorrelation"]
anc_sim["Metric"] = ["AdvancedNormalizedCorrelation"]
anc_nl["Metric"] = ["AdvancedNormalizedCorrelation"]

DEFAULT_REG_PARAM_MAPS["rigid_anc"] = anc_rig
DEFAULT_REG_PARAM_MAPS["affine_anc"] = anc_aff
DEFAULT_REG_PARAM_MAPS["similarity_anc"] = anc_sim
DEFAULT_REG_PARAM_MAPS["nl_anc"] = anc_nl
